#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The HRIT electrol reader tests package."""

import datetime
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest
from xarray import DataArray

from satpy.readers.electrol_hrit import (
    HRITGOMSEpilogueFileHandler,
    HRITGOMSFileHandler,
    HRITGOMSPrologueFileHandler,
    epilogue,
    image_acquisition,
    prologue,
    recarray2dict,
    satellite_status,
)
from satpy.tests.utils import make_dataid

# Simplify some type selections
f64_t = np.float64
i32_t = np.int32
u32_t = np.uint32


class Testrecarray2dict(unittest.TestCase):
    """Test the function that converts numpy record arrays into dicts for use within SatPy."""

    def test_fun(self):
        """Test record array."""
        inner_st = np.dtype([("test_str", "<S20"), ("test_int", "i4")])
        outer_st = np.dtype([("test_sec", inner_st), ("test_flt", "f4")])

        inner_da = np.array([("Testing", 10)], dtype=inner_st)
        outer_da = np.array([(inner_da, 1.45)], dtype=outer_st)

        expected = {"test_sec": {"test_str": np.array([b"Testing"], dtype="<S20"),
                                 "test_int": np.array([10], dtype=np.int32)},
                    "test_flt": np.array([1.45], dtype=np.float32)}
        assert expected == recarray2dict(outer_da)


class TestHRITGOMSProFileHandler(unittest.TestCase):
    """Test the HRIT Prologue FileHandler."""

    # Below are variable definitions used in testing the prologue reader.
    # These values are taken from a typical ELECTRO-L HRIT scene
    test_sat_status = {"TagType": 2,
                       "TagLength": 292,
                       "SatelliteID": 19002,
                       "SatelliteName": b"ELECTRO",
                       "NominalLongitude": 1.3264,
                       "SatelliteCondition": 1,
                       "TimeOffset": 0.}

    test_img_acq = {"TagType": np.repeat(3, 10).astype(u32_t),
                    "TagLength": np.repeat(24, 10).astype(u32_t),
                    "Status": np.repeat(2, 10).astype(u32_t),
                    "StartDelay": np.repeat(9119019, 10).astype(np.int32),
                    "Cel": np.repeat(0., 10)}

    test_calib = np.full((10, 1024), 50, dtype=np.int32)

    test_pro = {"SatelliteStatus": test_sat_status,
                "ImageAcquisition": test_img_acq,
                "ImageCalibration": test_calib}

    @mock.patch("satpy.readers.electrol_hrit.np.fromfile")
    @mock.patch("satpy.readers.electrol_hrit.HRITFileHandler.__init__")
    def test_init(self, new_fh_init, fromfile):
        """Set up the hrit file handler for testing."""
        new_fh_init.return_value.filename = "filename"
        HRITGOMSPrologueFileHandler.filename = "filename"
        HRITGOMSPrologueFileHandler.mda = {"total_header_length": 1}

        # Set up the test data to use within the prologue reader
        tss = np.array([(2, 292, 19002, "ELECTRO", 1.3264, 1, 0.)],
                       dtype=satellite_status)
        tia = np.tile(np.array([(3, 24, 2, 9119019, 0.)],
                      dtype=image_acquisition), (1, 10))
        tc = np.full((10, 1024), 50)
        rtv = np.array([(tss, tia, tc)], dtype=prologue)
        # Pretend to return this when reading a fake prologue
        fromfile.return_value = rtv
        m = mock.mock_open()

        with mock.patch("satpy.readers.electrol_hrit.open",
                        m, create=True) as newopen:
            newopen.return_value.__enter__.return_value.seek.return_value = 1
            test_t = datetime.datetime(2018, 1, 1, 0, 0)
            self.reader = HRITGOMSPrologueFileHandler(
                             "filename", {"platform_shortname": "GOMS2",
                                          "start_time": test_t,
                                          "service": "test_service"},
                             {"filetype": "info"})

        # assertDictEqual doesn't seem to work for dicts containing dicts,
        # so we must compare some items individually
        assert self.test_pro["SatelliteStatus"] == self.reader.prologue["SatelliteStatus"]
        prop = "ImageAcquisition"
        for key in self.reader.prologue[prop]:
            np.testing.assert_array_equal(self.test_pro[prop][key],
                                          self.reader.prologue[prop][key])
        np.testing.assert_array_equal(self.test_pro["ImageCalibration"],
                                      self.reader.prologue["ImageCalibration"])


class TestHRITGOMSEpiFileHandler(unittest.TestCase):
    """Test the HRIT Epilogue FileHandler."""

    @mock.patch("satpy.readers.electrol_hrit.np.fromfile")
    @mock.patch("satpy.readers.electrol_hrit.HRITFileHandler.__init__")
    def test_init(self, new_fh_init, fromfile):
        """Set up the hrit file handler for testing."""
        new_fh_init.return_value.filename = "filename"
        HRITGOMSEpilogueFileHandler.filename = "filename"
        HRITGOMSEpilogueFileHandler.mda = {"total_header_length": 1}

        # Set up the test data to use within the epilogue reader
        rtv = np.ones((1,), dtype=epilogue)
        # Pretend to return this when reading a fake epilogue
        fromfile.return_value = rtv
        m = mock.mock_open()

        with mock.patch("satpy.readers.electrol_hrit.open",
                        m, create=True) as newopen:
            newopen.return_value.__enter__.return_value.seek.return_value = 1
            test_t = datetime.datetime(2018, 1, 1, 0, 0)
            self.reader = HRITGOMSEpilogueFileHandler(
                             "filename", {"platform_shortname": "GOMS2",
                                          "start_time": test_t,
                                          "service": "test_service"},
                             {"filetype": "info"})

            epi = self.reader.epilogue

            # We don't check everything in the epilogue (too many nested dicts)
            # but rather check the epilogue is returned as a dict and that two
            # representative data fields are as we expect.
            assert isinstance(epi, dict)
            np.testing.assert_array_equal(
                epi["RadiometricProcessing"]["RPSummary"]["IsOptic"], np.ones(10))
            np.testing.assert_array_equal(
                epi["GeometricProcessing"]["TimeProcessing"], np.ones(10))


@mock.patch("satpy.readers.electrol_hrit.HRITGOMSFileHandler.__init__", return_value=None)
@mock.patch("satpy.readers.electrol_hrit.HRITFileHandler.get_dataset", return_value={})
class TestHRITGOMSFileHandler(unittest.TestCase):
    """A test of the ELECTRO-L main file handler functions."""

    @mock.patch("satpy.readers.electrol_hrit.HRITGOMSFileHandler.calibrate")
    def test_get_dataset(self, calibrate_mock, *mocks):
        """Test get dataset."""
        key = make_dataid(name="CH1", calibration="counts")
        fake_array = mock.MagicMock()
        fake_array.attrs = dict()
        calibrate_mock.return_value = fake_array
        fh = HRITGOMSFileHandler()
        fh.platform_name = "Electro"
        fh.mda = {"projection_parameters": {"SSP_longitude": 0.0},
                  "orbital_parameters": {"satellite_nominal_longitude": 0.5}}
        info = {"units": "m", "standard_name": "electro", "wavelength": 5.0}
        output = fh.get_dataset(key, info)

        # Check that 'calibrate' is called
        calibrate_mock.assert_called()

        # Check that the correct attributes are returned
        attrs_exp = info.copy()
        attrs_exp.update({"orbital_parameters": {"satellite_nominal_longitude": 0.5,
                                                 "satellite_nominal_latitude": 0.0,
                                                 "projection_longitude": 0.0,
                                                 "projection_latitude": 0.0,
                                                 "projection_altitude": 35785831.00},
                          "platform_name": "Electro",
                          "sensor": "msu-gs"})
        assert dict(output.attrs, **attrs_exp) == output.attrs

    def test_calibrate(self, *mocks):
        """Test calibrate."""
        lut = np.linspace(1e6, 1.6e6, num=1024).astype(np.int32)
        lut = np.tile(lut, (10, 1))
        fh = HRITGOMSFileHandler()
        fh.prologue = {"ImageCalibration": lut}
        fh.chid = 1

        # Set up test input data

        counts = DataArray(da.linspace(1, 1023, 25, chunks=5,
                                       dtype=np.uint16).reshape(5, 5))

        # Test that calibration fails if given a silly mode
        with pytest.raises(NotImplementedError):
            fh.calibrate(counts, "nonsense")

        # Test that 'counts' calibration returns identical values to input
        out = fh.calibrate(counts, "counts")
        assert np.all(out.values == counts.values)

        # Test that 'radiance' calibrates successfully
        out = fh.calibrate(counts, "radiance")
        assert np.allclose(out.values, lut[0, counts] / 1000.0)

        # Test that 'brightness_temperature' calibrates successfully
        out = fh.calibrate(counts, "brightness_temperature")
        assert np.allclose(out.values, lut[0, counts] / 1000.0)

    def test_get_area_def(self, *mocks):
        """Test get_area_def."""
        example_area_ext = (-5566748.0802, -1854249.1809,
                            5570748.6178, 2000.2688)

        fh = HRITGOMSFileHandler()
        fh.mda = {"cfac": 10231753, "lfac": 10231753,
                  "coff": 1392.0, "loff": 0.0, "number_of_lines": 464,
                  "number_of_columns": 2784,
                  "projection_parameters": {"SSP_longitude": 0.0}}
        area = fh.get_area_def(True)

        assert np.allclose(np.array(area.area_extent), np.array(example_area_ext))



"""Test the flash age compositor."""
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.composites.lightning import LightningTimeCompositor


def test_flash_age_compositor():
    """Test the flash_age compsitor by comparing two xarrays object."""
    comp = LightningTimeCompositor("flash_age",prerequisites=["flash_time"],
                                   standard_name="lightning_time",
                                   time_range=60,
                                   reference_time="end_time")
    attrs_flash_age = {"variable_name": "flash_time","name": "flash_time",
                       "start_time": datetime.datetime(2024, 8, 1, 10, 50, 0),
                       "end_time": datetime.datetime(2024, 8, 1, 11, 0, 0),"reader": "li_l2_nc"}
    flash_age_value = da.array(["2024-08-01T09:00:00",
            "2024-08-01T10:00:00", "2024-08-01T10:30:00","2024-08-01T11:00:00"], dtype="datetime64[ns]")
    flash_age = xr.DataArray(
        flash_age_value,
        dims=["y"],
        coords={
            "crs": "8B +proj=longlat +ellps=WGS84 +type=crs"},
        attrs = attrs_flash_age,
        name="flash_time")
    res = comp([flash_age])
    expected_attrs = {"variable_name": "flash_time","name": "lightning_time",
                       "start_time": datetime.datetime(2024, 8, 1, 10, 50, 0),
                       "end_time": datetime.datetime(2024, 8, 1, 11, 0, 0),"reader": "li_l2_nc",
                       "standard_name": "lightning_time"
                       }
    expected_array = xr.DataArray(da.array([np.nan, 0.0,0.5,1.0]),
                                  dims=["y"],
                                  coords={
                                      "crs": "8B +proj=longlat +ellps=WGS84 +type=crs"},
                                  attrs = expected_attrs,
                                  name="flash_time")
    xr.testing.assert_equal(res,expected_array)

def test_empty_array_error(caplog):
    """Test when the filtered array is empty."""
    comp = LightningTimeCompositor("flash_age",prerequisites=["flash_time"],
                                   standard_name="lightning_time",
                                   time_range=60,
                                   reference_time="end_time")
    attrs_flash_age = {"variable_name": "flash_time","name": "flash_time",
                       "start_time": np.datetime64(datetime.datetime(2024, 8, 1, 10, 0, 0)),
                       "end_time": datetime.datetime(2024, 8, 1, 11, 0, 0),
                       "reader": "li_l2_nc"}
    flash_age_value = da.array(["2024-08-01T09:00:00"], dtype="datetime64[ns]")
    flash_age = xr.DataArray(flash_age_value,
                             dims=["y"],
                             coords={
                                 "crs": "8B +proj=longlat +ellps=WGS84 +type=crs"},
                             attrs = attrs_flash_age,
                             name="flash_time")
    with caplog.at_level(logging.WARNING):
        _ = comp([flash_age])
    # Assert that the log contains the expected warning message
    assert "All the flash_age events happened before" in caplog.text

def test_update_missing_metadata():
    """Test the _update_missing_metadata method."""
    existing_attrs = {
        "standard_name": "lightning_event_time",
        "time_range": 30
    }

    # New metadata to be merged
    new_attrs = {
        "standard_name": None,  # Should not overwrite since it's None
        "reference_time": "2023-09-20T00:00:00Z",  # Should be added
        "units": "seconds"  # Should be added
    }

    # Expected result after merging
    expected_attrs = {
        "standard_name": "lightning_event_time",  # Should remain the same
        "time_range": 30,  # Should remain the same
        "reference_time": "2023-09-20T00:00:00Z",  # Should be added
        "units": "seconds"  # Should be added
    }

    # Call the static method
    LightningTimeCompositor._update_missing_metadata(existing_attrs, new_attrs)

    # Assert the final state of existing_attrs is as expected
    assert existing_attrs == expected_attrs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Testing the yaml_reader module."""

import datetime as dt
import os
import random
import unittest
from tempfile import mkdtemp
from unittest.mock import MagicMock, call, patch

import dask.array as da
import numpy as np
import pytest
import xarray as xr

import satpy.readers.core.yaml_reader as yr
from satpy._compat import cache
from satpy.dataset import DataQuery
from satpy.dataset.dataid import ModifierTuple
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.pmw_channels_definitions import FrequencyDoubleSideBand, FrequencyRange
from satpy.tests.utils import make_dataid

MHS_YAML_READER_DICT = {
    "reader": {"name": "mhs_l1c_aapp",
               "description": "AAPP l1c Reader for AMSU-B/MHS data",
               "sensors": ["mhs"],
               "data_identification_keys": {"name": {"required": True},
                                            "frequency_double_sideband":
                                                {"type": FrequencyDoubleSideBand},
                                            "frequency_range": {"type": FrequencyRange},
                                            "resolution": None,
                                            "polarization": {"enum": ["H", "V"]},
                                            "calibration": {"enum": ["brightness_temperature"], "transitive": True},
                                            "modifiers": {"required": True,
                                                          "default": [],
                                                          "type": ModifierTuple}},
               "config_files": ("satpy/etc/readers/mhs_l1c_aapp.yaml",)},
    "datasets": {"1": {"name": "1",
                       "frequency_range": {"central": 89.0, "bandwidth": 2.8, "unit": "GHz"},
                       "polarization": "V",
                       "resolution": 16000,
                       "calibration": {"brightness_temperature": {"standard_name": "toa_brightness_temperature"}},
                       "coordinates": ["longitude", "latitude"],
                       "file_type": "mhs_aapp_l1c"},
                 "2": {"name": "2",
                       "frequency_range": {"central": 157.0, "bandwidth": 2.8, "unit": "GHz"},
                       "polarization": "V",
                       "resolution": 16000,
                       "calibration": {"brightness_temperature": {"standard_name": "toa_brightness_temperature"}},
                       "coordinates": ["longitude", "latitude"],
                       "file_type": "mhs_aapp_l1c"},
                 "3": {"name": "3",
                       "frequency_double_sideband": {"unit": "GHz",
                                                     "central": 183.31,
                                                     "side": 1.0,
                                                     "bandwidth": 1.0},
                       "polarization": "V",
                       "resolution": 16000,
                       "calibration": {"brightness_temperature": {"standard_name": "toa_brightness_temperature"}},
                       "coordinates": ["longitude", "latitude"],
                       "file_type": "mhs_aapp_l1c"}},
    "file_types": {"mhs_aapp_l1c": {"file_reader": BaseFileHandler,
                                    "file_patterns": [
                                        'mhsl1c_{platform_shortname}_{start_time:%Y%m%d_%H%M}_{orbit_number:05d}.l1c']}}}  # noqa


class FakeFH(BaseFileHandler):
    """Fake file handler class."""

    def __init__(self, start_time, end_time):
        """Initialize fake file handler."""
        super(FakeFH, self).__init__("", {}, {})
        self._start_time = start_time
        self._end_time = end_time
        self.get_bounding_box = MagicMock()
        fake_ds = MagicMock()
        fake_ds.return_value.dims = ["x", "y"]
        self.get_dataset = fake_ds
        self.combine_info = MagicMock()

    @property
    def start_time(self):
        """Return start time."""
        return self._start_time

    @property
    def end_time(self):
        """Return end time."""
        return self._end_time


class TestUtils(unittest.TestCase):
    """Test the utility functions."""

    def test_get_filebase(self):
        """Check the get_filebase function."""
        base_dir = os.path.join(os.path.expanduser("~"), "data",
                                "satellite", "Sentinel-3")
        base_data = ("S3A_OL_1_EFR____20161020T081224_20161020T081524_"
                     "20161020T102406_0179_010_078_2340_SVL_O_NR_002.SEN3")
        base_dir = os.path.join(base_dir, base_data)
        pattern = ("{mission_id:3s}_OL_{processing_level:1s}_{datatype_id:_<6s"
                   "}_{start_time:%Y%m%dT%H%M%S}_{end_time:%Y%m%dT%H%M%S}_{cre"
                   "ation_time:%Y%m%dT%H%M%S}_{duration:4d}_{cycle:3d}_{relati"
                   "ve_orbit:3d}_{frame:4d}_{centre:3s}_{mode:1s}_{timeliness:"
                   "2s}_{collection:3s}.SEN3/geo_coordinates.nc")
        pattern = os.path.join(*pattern.split("/"))
        filename = os.path.join(base_dir, "Oa05_radiance.nc")
        expected = os.path.join(base_data, "Oa05_radiance.nc")
        assert yr._get_filebase(filename, pattern) == expected

    def test_match_filenames(self):
        """Check that matching filenames works."""
        # just a fake path for testing that doesn't have to exist
        base_dir = os.path.join(os.path.expanduser("~"), "data",
                                "satellite", "Sentinel-3")
        base_data = ("S3A_OL_1_EFR____20161020T081224_20161020T081524_"
                     "20161020T102406_0179_010_078_2340_SVL_O_NR_002.SEN3")
        base_dir = os.path.join(base_dir, base_data)
        pattern = ("{mission_id:3s}_OL_{processing_level:1s}_{datatype_id:_<6s"
                   "}_{start_time:%Y%m%dT%H%M%S}_{end_time:%Y%m%dT%H%M%S}_{cre"
                   "ation_time:%Y%m%dT%H%M%S}_{duration:4d}_{cycle:3d}_{relati"
                   "ve_orbit:3d}_{frame:4d}_{centre:3s}_{mode:1s}_{timeliness:"
                   "2s}_{collection:3s}.SEN3/geo_coordinates.nc")
        pattern = os.path.join(*pattern.split("/"))
        filenames = [os.path.join(base_dir, "Oa05_radiance.nc"),
                     os.path.join(base_dir, "geo_coordinates.nc")]
        expected = os.path.join(base_dir, "geo_coordinates.nc")
        assert yr._match_filenames(filenames, pattern) == {expected}

    def test_match_filenames_windows_forward_slash(self):
        """Check that matching filenames works on Windows with forward slashes.

        This is common from Qt5 which internally uses forward slashes everywhere.

        """
        # just a fake path for testing that doesn't have to exist
        base_dir = os.path.join(os.path.expanduser("~"), "data",
                                "satellite", "Sentinel-3")
        base_data = ("S3A_OL_1_EFR____20161020T081224_20161020T081524_"
                     "20161020T102406_0179_010_078_2340_SVL_O_NR_002.SEN3")
        base_dir = os.path.join(base_dir, base_data)
        pattern = ("{mission_id:3s}_OL_{processing_level:1s}_{datatype_id:_<6s"
                   "}_{start_time:%Y%m%dT%H%M%S}_{end_time:%Y%m%dT%H%M%S}_{cre"
                   "ation_time:%Y%m%dT%H%M%S}_{duration:4d}_{cycle:3d}_{relati"
                   "ve_orbit:3d}_{frame:4d}_{centre:3s}_{mode:1s}_{timeliness:"
                   "2s}_{collection:3s}.SEN3/geo_coordinates.nc")
        pattern = os.path.join(*pattern.split("/"))
        filenames = [os.path.join(base_dir, "Oa05_radiance.nc").replace(os.sep, "/"),
                     os.path.join(base_dir, "geo_coordinates.nc").replace(os.sep, "/")]
        expected = os.path.join(base_dir, "geo_coordinates.nc").replace(os.sep, "/")
        assert yr._match_filenames(filenames, pattern) == {expected}

    def test_listify_string(self):
        """Check listify_string."""
        assert yr.listify_string(None) == []
        assert yr.listify_string("some string") == ["some string"]
        assert yr.listify_string(["some", "string"]) == ["some", "string"]


class DummyReader(BaseFileHandler):
    """Dummy reader instance."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the dummy reader."""
        super(DummyReader, self).__init__(
            filename, filename_info, filetype_info)
        self._start_time = dt.datetime(2000, 1, 1, 12, 1)
        self._end_time = dt.datetime(2000, 1, 1, 12, 2)
        self.metadata = {}

    @property
    def start_time(self):
        """Return start time."""
        return self._start_time

    @property
    def end_time(self):
        """Return end time."""
        return self._end_time


class TestFileFileYAMLReaderMultiplePatterns(unittest.TestCase):
    """Test units from FileYAMLReader with multiple readers."""

    def setUp(self):
        """Prepare a reader instance with a fake config."""
        patterns = ["a{something:3s}.bla",
                    "a0{something:2s}.bla"]
        res_dict = {"reader": {"name": "fake",
                               "sensors": ["canon"]},
                    "file_types": {"ftype1": {"name": "ft1",
                                              "file_patterns": patterns,
                                              "file_reader": DummyReader}},
                    "datasets": {"ch1": {"name": "ch01",
                                         "wavelength": [0.5, 0.6, 0.7],
                                         "calibration": "reflectance",
                                         "file_type": "ftype1",
                                         "coordinates": ["lons", "lats"]},
                                 "ch2": {"name": "ch02",
                                         "wavelength": [0.7, 0.75, 0.8],
                                         "calibration": "counts",
                                         "file_type": "ftype1",
                                         "coordinates": ["lons", "lats"]},
                                 "lons": {"name": "lons",
                                          "file_type": "ftype2"},
                                 "lats": {"name": "lats",
                                          "file_type": "ftype2"}}}

        self.config = res_dict
        self.reader = yr.FileYAMLReader(self.config,
                                        filter_parameters={
                                            "start_time": dt.datetime(2000, 1, 1),
                                            "end_time": dt.datetime(2000, 1, 2)})

    def test_select_from_pathnames(self):
        """Check select_files_from_pathnames."""
        filelist = ["a001.bla", "a002.bla", "abcd.bla", "k001.bla", "a003.bli"]

        res = self.reader.select_files_from_pathnames(filelist)
        for expected in ["a001.bla", "a002.bla", "abcd.bla"]:
            assert expected in res
        assert len(res) == 3

    def test_fn_items_for_ft(self):
        """Check filename_items_for_filetype."""
        filelist = ["a001.bla", "a002.bla", "abcd.bla", "k001.bla", "a003.bli"]
        ft_info = self.config["file_types"]["ftype1"]
        fiter = self.reader.filename_items_for_filetype(filelist, ft_info)

        filenames = dict(fname for fname in fiter)
        assert len(filenames.keys()) == 3

    def test_create_filehandlers(self):
        """Check create_filehandlers."""
        filelist = ["a001.bla", "a002.bla", "a001.bla", "a002.bla",
                    "abcd.bla", "k001.bla", "a003.bli"]

        self.reader.create_filehandlers(filelist)
        assert len(self.reader.file_handlers["ftype1"]) == 3

    def test_serializable(self):
        """Check that a reader is serializable by dask.

        This ensures users are able to serialize a Scene object that contains
        readers.
        """
        from distributed.protocol import deserialize, serialize
        filelist = ["a001.bla", "a002.bla", "a001.bla", "a002.bla",
                    "abcd.bla", "k001.bla", "a003.bli"]

        self.reader.create_filehandlers(filelist)
        cloned_reader = deserialize(*serialize(self.reader))
        assert self.reader.file_handlers.keys() == cloned_reader.file_handlers.keys()
        assert self.reader.all_ids == cloned_reader.all_ids


class TestFileYAMLReaderWithCustomIDKey(unittest.TestCase):
    """Test units from FileYAMLReader with custom id_keys."""

    def setUp(self):
        """Set up the test case."""
        self.config = MHS_YAML_READER_DICT
        self.reader = yr.FileYAMLReader(MHS_YAML_READER_DICT,
                                        filter_parameters={
                                            "start_time": dt.datetime(2000, 1, 1),
                                            "end_time": dt.datetime(2000, 1, 2),
                                        })

    def test_custom_type_with_dict_contents_gets_parsed_correctly(self):
        """Test custom type with dictionary contents gets parsed correctly."""
        ds_ids = list(self.reader.all_dataset_ids)
        assert ds_ids[0]["frequency_range"] == FrequencyRange(89., 2.8, "GHz")

        assert ds_ids[2]["frequency_double_sideband"] == FrequencyDoubleSideBand(183.31, 1., 1., "GHz")


class TestFileFileYAMLReader(unittest.TestCase):
    """Test units from FileYAMLReader."""

    def setUp(self):
        """Prepare a reader instance with a fake config."""
        patterns = ["a{something:3s}.bla"]
        res_dict = {"reader": {"name": "fake",
                               "sensors": ["canon"]},
                    "file_types": {"ftype1": {"name": "ft1",
                                              "file_reader": BaseFileHandler,
                                              "file_patterns": patterns}},
                    "datasets": {"ch1": {"name": "ch01",
                                         "wavelength": [0.5, 0.6, 0.7],
                                         "calibration": "reflectance",
                                         "file_type": "ftype1",
                                         "coordinates": ["lons", "lats"]},
                                 "ch2": {"name": "ch02",
                                         "wavelength": [0.7, 0.75, 0.8],
                                         "calibration": "counts",
                                         "file_type": "ftype1",
                                         "coordinates": ["lons", "lats"]},
                                 "lons": {"name": "lons",
                                          "file_type": "ftype2"},
                                 "lats": {"name": "lats",
                                          "file_type": "ftype2"}}}

        self.config = res_dict
        self.reader = yr.FileYAMLReader(res_dict,
                                        filter_parameters={
                                            "start_time": dt.datetime(2000, 1, 1),
                                            "end_time": dt.datetime(2000, 1, 2),
                                        })

    def test_deprecated_passing_config_files(self):
        """Test that we get an exception when config files are passed to inti."""
        with pytest.raises(ValueError,
                           match="Passing config files to create a Reader is deprecated.*"):
            yr.FileYAMLReader("/path/to/some/file.yaml")

    def test_all_data_ids(self):
        """Check that all datasets ids are returned."""
        for dataid in self.reader.all_dataset_ids:
            name = dataid["name"].replace("0", "")
            assert self.config["datasets"][name]["name"] == dataid["name"]
            if "wavelength" in self.config["datasets"][name]:
                assert self.config["datasets"][name]["wavelength"] == list(dataid["wavelength"])[:3]
            if "calibration" in self.config["datasets"][name]:
                assert self.config["datasets"][name]["calibration"] == dataid["calibration"]

    def test_all_dataset_names(self):
        """Get all dataset names."""
        assert self.reader.all_dataset_names == set(["ch01", "ch02", "lons", "lats"])

    def test_available_dataset_ids(self):
        """Get ids of the available datasets."""
        loadables = self.reader.select_files_from_pathnames(["a001.bla"])
        self.reader.create_filehandlers(loadables)
        assert set(self.reader.available_dataset_ids) == {make_dataid(name="ch02", wavelength=(0.7, 0.75, 0.8),
                                                                      calibration="counts", modifiers=()),
                                                          make_dataid(name="ch01", wavelength=(0.5, 0.6, 0.7),
                                                                      calibration="reflectance", modifiers=())}

    def test_available_dataset_names(self):
        """Get ids of the available datasets."""
        loadables = self.reader.select_files_from_pathnames(["a001.bla"])
        self.reader.create_filehandlers(loadables)
        assert set(self.reader.available_dataset_names) == set(["ch01", "ch02"])

    def test_filter_fh_by_time(self):
        """Check filtering filehandlers by time."""
        fh0 = FakeFH(dt.datetime(1999, 12, 30),
                     dt.datetime(1999, 12, 31))
        fh1 = FakeFH(dt.datetime(1999, 12, 31, 10, 0),
                     dt.datetime(2000, 1, 1, 12, 30))
        fh2 = FakeFH(dt.datetime(2000, 1, 1, 10, 0),
                     dt.datetime(2000, 1, 1, 12, 30))
        fh3 = FakeFH(dt.datetime(2000, 1, 1, 12, 30),
                     dt.datetime(2000, 1, 2, 12, 30))
        fh4 = FakeFH(dt.datetime(2000, 1, 2, 12, 30),
                     dt.datetime(2000, 1, 3, 12, 30))
        fh5 = FakeFH(dt.datetime(1999, 12, 31, 10, 0),
                     dt.datetime(2000, 1, 3, 12, 30))

        for idx, fh in enumerate([fh0, fh1, fh2, fh3, fh4, fh5]):
            res = self.reader.time_matches(fh.start_time, fh.end_time)
            # only the first one should be false
            assert res == (idx not in [0, 4])

        for idx, fh in enumerate([fh0, fh1, fh2, fh3, fh4, fh5]):
            res = self.reader.time_matches(fh.start_time, None)
            assert res == (idx not in [0, 1, 4, 5])

    @patch("satpy.readers.core.yaml_reader.get_area_def")
    @patch("satpy.readers.core.yaml_reader.AreaDefBoundary")
    @patch("satpy.readers.core.yaml_reader.Boundary")
    def test_file_covers_area(self, bnd, adb, gad):
        """Test that area coverage is checked properly."""
        file_handler = FakeFH(dt.datetime(1999, 12, 31, 10, 0),
                              dt.datetime(2000, 1, 3, 12, 30))

        self.reader.filter_parameters["area"] = True
        bnd.return_value.contour_poly.intersection.return_value = True
        adb.return_value.contour_poly.intersection.return_value = True
        res = self.reader.check_file_covers_area(file_handler, True)
        assert res

        bnd.return_value.contour_poly.intersection.return_value = False
        adb.return_value.contour_poly.intersection.return_value = False
        res = self.reader.check_file_covers_area(file_handler, True)
        assert not res

        file_handler.get_bounding_box.side_effect = NotImplementedError()
        self.reader.filter_parameters["area"] = True
        res = self.reader.check_file_covers_area(file_handler, True)
        assert res

    def test_start_end_time(self):
        """Check start and end time behaviours."""
        self.reader.file_handlers = {}

        with pytest.raises(RuntimeError):
            self.reader.start_time

        with pytest.raises(RuntimeError):
            self.reader.end_time

        fh0 = FakeFH(dt.datetime(1999, 12, 30, 0, 0),
                     dt.datetime(1999, 12, 31, 0, 0))
        fh1 = FakeFH(dt.datetime(1999, 12, 31, 10, 0),
                     dt.datetime(2000, 1, 1, 12, 30))
        fh2 = FakeFH(dt.datetime(2000, 1, 1, 10, 0),
                     dt.datetime(2000, 1, 1, 12, 30))
        fh3 = FakeFH(dt.datetime(2000, 1, 1, 12, 30),
                     dt.datetime(2000, 1, 2, 12, 30))
        fh4 = FakeFH(dt.datetime(2000, 1, 2, 12, 30),
                     dt.datetime(2000, 1, 3, 12, 30))
        fh5 = FakeFH(dt.datetime(1999, 12, 31, 10, 0),
                     dt.datetime(2000, 1, 3, 12, 30))

        self.reader.file_handlers = {
            "0": [fh1, fh2, fh3, fh4, fh5],
            "1": [fh0, fh1, fh2, fh3, fh4, fh5],
            "2": [fh2, fh3],
        }

        assert self.reader.start_time == dt.datetime(1999, 12, 30, 0, 0)
        assert self.reader.end_time == dt.datetime(2000, 1, 3, 12, 30)

    def test_select_from_pathnames(self):
        """Check select_files_from_pathnames."""
        filelist = ["a001.bla", "a002.bla", "abcd.bla", "k001.bla", "a003.bli"]

        res = self.reader.select_files_from_pathnames(filelist)
        for expected in ["a001.bla", "a002.bla", "abcd.bla"]:
            assert expected in res

        assert 0 == len(self.reader.select_files_from_pathnames([]))

    def test_select_from_directory(self):
        """Check select_files_from_directory."""
        filelist = ["a001.bla", "a002.bla", "abcd.bla", "k001.bla", "a003.bli"]
        dpath = mkdtemp()
        for fname in filelist:
            with open(os.path.join(dpath, fname), "w"):
                pass

        res = self.reader.select_files_from_directory(dpath)
        for expected in ["a001.bla", "a002.bla", "abcd.bla"]:
            assert os.path.join(dpath, expected) in res

        for fname in filelist:
            os.remove(os.path.join(dpath, fname))
        assert 0 == len(self.reader.select_files_from_directory(dpath))
        os.rmdir(dpath)

        from fsspec.implementations.local import LocalFileSystem

        class Silly(LocalFileSystem):
            def glob(self, pattern):
                return ["/grocery/apricot.nc", "/grocery/aubergine.nc"]

        res = self.reader.select_files_from_directory(dpath, fs=Silly())
        assert res == {"/grocery/apricot.nc", "/grocery/aubergine.nc"}

    def test_supports_sensor(self):
        """Check supports_sensor."""
        assert self.reader.supports_sensor("canon")
        assert not self.reader.supports_sensor("nikon")

    @patch("satpy.readers.core.yaml_reader.StackedAreaDefinition")
    def test_load_area_def(self, sad):
        """Test loading the area def for the reader."""
        dataid = MagicMock()
        file_handlers = []
        items = random.randrange(2, 10)
        for _i in range(items):
            file_handlers.append(MagicMock())
        final_area = self.reader._load_area_def(dataid, file_handlers)
        assert final_area == sad.return_value.squeeze.return_value

        args, kwargs = sad.call_args
        assert len(args) == items

    def test_preferred_filetype(self):
        """Test finding the preferred filetype."""
        self.reader.file_handlers = {"a": "a", "b": "b", "c": "c"}
        assert self.reader._preferred_filetype(["c", "a"]) == "c"
        assert self.reader._preferred_filetype(["a", "c"]) == "a"
        assert self.reader._preferred_filetype(["d", "e"]) is None

    def test_get_coordinates_for_dataset_key(self):
        """Test getting coordinates for a key."""
        ds_q = DataQuery(name="ch01", wavelength=(0.5, 0.6, 0.7, "µm"),
                         calibration="reflectance", modifiers=())
        res = self.reader._get_coordinates_for_dataset_key(ds_q)
        assert res == [make_dataid(name="lons"), make_dataid(name="lats")]

    def test_get_coordinates_for_dataset_key_without(self):
        """Test getting coordinates for a key without coordinates."""
        ds_id = make_dataid(name="lons",
                            modifiers=())
        res = self.reader._get_coordinates_for_dataset_key(ds_id)
        assert res == []

    def test_get_coordinates_for_dataset_keys(self):
        """Test getting coordinates for keys."""
        ds_id1 = make_dataid(name="ch01", wavelength=(0.5, 0.6, 0.7),
                             calibration="reflectance", modifiers=())
        ds_id2 = make_dataid(name="ch02", wavelength=(0.7, 0.75, 0.8),
                             calibration="counts", modifiers=())
        lons = make_dataid(name="lons", modifiers=())
        lats = make_dataid(name="lats", modifiers=())

        res = self.reader._get_coordinates_for_dataset_keys([ds_id1, ds_id2,
                                                             lons])
        expected = {ds_id1: [lons, lats], ds_id2: [lons, lats], lons: []}

        assert res == expected

    def test_get_file_handlers(self):
        """Test getting filehandler to load a dataset."""
        ds_id1 = make_dataid(name="ch01", wavelength=(0.5, 0.6, 0.7),
                             calibration="reflectance", modifiers=())
        self.reader.file_handlers = {"ftype1": "bla"}

        assert self.reader._get_file_handlers(ds_id1) == "bla"

        lons = make_dataid(name="lons", modifiers=())
        assert self.reader._get_file_handlers(lons) is None

    @patch("satpy.readers.core.yaml_reader.xr")
    def test_load_entire_dataset(self, xarray):
        """Check loading an entire dataset."""
        file_handlers = [FakeFH(None, None), FakeFH(None, None),
                         FakeFH(None, None), FakeFH(None, None)]

        proj = self.reader._load_dataset(None, {}, file_handlers)

        assert proj is xarray.concat.return_value


class TestFileYAMLReaderLoading(unittest.TestCase):
    """Tests for FileYAMLReader.load."""

    def setUp(self):
        """Prepare a reader instance with a fake config."""
        patterns = ["a{something:3s}.bla"]
        res_dict = {"reader": {"name": "fake",
                               "sensors": ["canon"]},
                    "file_types": {"ftype1": {"name": "ft1",
                                              "file_reader": BaseFileHandler,
                                              "file_patterns": patterns}},
                    "datasets": {"ch1": {"name": "ch01",
                                         "wavelength": [0.5, 0.6, 0.7],
                                         "calibration": "reflectance",
                                         "file_type": "ftype1"},
                                 }}

        self.config = res_dict
        self.reader = yr.FileYAMLReader(res_dict,
                                        filter_parameters={
                                            "start_time": dt.datetime(2000, 1, 1),
                                            "end_time": dt.datetime(2000, 1, 2),
                                        })
        fake_fh = FakeFH(None, None)
        self.lons = xr.DataArray(np.ones((2, 2)) * 2,
                                 dims=["y", "x"],
                                 attrs={"standard_name": "longitude",
                                        "name": "longitude"})
        self.lats = xr.DataArray(np.ones((2, 2)) * 2,
                                 dims=["y", "x"],
                                 attrs={"standard_name": "latitude",
                                        "name": "latitude"})
        self.data = None

        def _assign_array(dsid, *_args, **_kwargs):
            if dsid["name"] == "longitude":
                return self.lons
            if dsid["name"] == "latitude":
                return self.lats
            return self.data

        fake_fh.get_dataset.side_effect = _assign_array
        self.reader.file_handlers = {"ftype1": [fake_fh]}

    def test_load_dataset_with_builtin_coords(self):
        """Test loading a dataset with builtin coordinates."""
        self.data = xr.DataArray(np.ones((2, 2)),
                                 coords={"longitude": self.lons,
                                         "latitude": self.lats},
                                 dims=["y", "x"])

        self._check_area_for_ch01()

    def test_load_dataset_with_builtin_coords_in_wrong_order(self):
        """Test loading a dataset with builtin coordinates in the wrong order."""
        self.data = xr.DataArray(np.ones((2, 2)),
                                 coords={"latitude": self.lats,
                                         "longitude": self.lons},
                                 dims=["y", "x"])

        self._check_area_for_ch01()

    def _check_area_for_ch01(self):
        res = self.reader.load(["ch01"])
        assert "area" in res["ch01"].attrs
        np.testing.assert_array_equal(res["ch01"].attrs["area"].lons, self.lons)
        np.testing.assert_array_equal(res["ch01"].attrs["area"].lats, self.lats)
        assert res["ch01"].attrs.get("reader") == "fake"


class TestFileFileYAMLReaderMultipleFileTypes(unittest.TestCase):
    """Test units from FileYAMLReader with multiple file types."""

    def setUp(self):
        """Prepare a reader instance with a fake config."""
        # Example: GOES netCDF data
        #   a) From NOAA CLASS: ftype1, including coordinates
        #   b) From EUMETSAT: ftype2, coordinates in extra file (ftype3)
        #
        # For test completeness add one channel (ch3) which is only available
        # in ftype1.
        patterns1 = ["a.nc"]
        patterns2 = ["b.nc"]
        patterns3 = ["geo.nc"]
        res_dict = {"reader": {"name": "fake",
                               "sensors": ["canon"]},
                    "file_types": {"ftype1": {"name": "ft1",
                                              "file_patterns": patterns1},
                                   "ftype2": {"name": "ft2",
                                              "file_patterns": patterns2},
                                   "ftype3": {"name": "ft3",
                                              "file_patterns": patterns3}},
                    "datasets": {"ch1": {"name": "ch01",
                                         "wavelength": [0.5, 0.6, 0.7],
                                         "calibration": "reflectance",
                                         "file_type": ["ftype1", "ftype2"],
                                         "coordinates": ["lons", "lats"]},
                                 "ch2": {"name": "ch02",
                                         "wavelength": [0.7, 0.75, 0.8],
                                         "calibration": "counts",
                                         "file_type": ["ftype1", "ftype2"],
                                         "coordinates": ["lons", "lats"]},
                                 "ch3": {"name": "ch03",
                                         "wavelength": [0.8, 0.85, 0.9],
                                         "calibration": "counts",
                                         "file_type": "ftype1",
                                         "coordinates": ["lons", "lats"]},
                                 "lons": {"name": "lons",
                                          "file_type": ["ftype1", "ftype3"]},
                                 "lats": {"name": "lats",
                                          "file_type": ["ftype1", "ftype3"]}}}

        self.config = res_dict
        self.reader = yr.FileYAMLReader(self.config)

    def test_update_ds_ids_from_file_handlers(self):
        """Test updating existing dataset IDs with information from the file."""
        from functools import partial
        orig_ids = self.reader.all_ids

        for ftype, resol in zip(("ftype1", "ftype2"), (1, 2)):
            # need to copy this because the dataset infos will be modified
            _orig_ids = {key: val.copy() for key, val in orig_ids.items()}
            with patch.dict(self.reader.all_ids, _orig_ids, clear=True), \
                    patch.dict(self.reader.available_ids, {}, clear=True):
                # Add a file handler with resolution property
                fh = MagicMock(filetype_info={"file_type": ftype},
                               resolution=resol)
                fh.available_datasets = partial(available_datasets, fh)
                fh.file_type_matches = partial(file_type_matches, fh)
                self.reader.file_handlers = {
                    ftype: [fh]}

                # Update existing dataset IDs with resolution property from
                # the file handler
                self.reader.update_ds_ids_from_file_handlers()

                # Make sure the resolution property has been transferred
                # correctly from the file handler to the dataset ID
                for ds_id, ds_info in self.reader.all_ids.items():
                    file_types = ds_info["file_type"]
                    if not isinstance(file_types, list):
                        file_types = [file_types]
                    if ftype in file_types:
                        assert resol == ds_id["resolution"]


# Test methods


def available_datasets(self, configured_datasets=None):
    """Fake available_datasets for testing multiple file types."""
    res = self.resolution
    # update previously configured datasets
    for is_avail, ds_info in (configured_datasets or []):
        if is_avail is not None:
            yield is_avail, ds_info

        matches = self.file_type_matches(ds_info["file_type"])
        if matches and ds_info.get("resolution") != res:
            new_info = ds_info.copy()
            new_info["resolution"] = res
            yield True, new_info
        elif is_avail is None:
            yield is_avail, ds_info


def file_type_matches(self, ds_ftype):
    """Fake file_type_matches for testing multiple file types."""
    if isinstance(ds_ftype, str) and ds_ftype == self.filetype_info["file_type"]:
        return True
    if self.filetype_info["file_type"] in ds_ftype:
        return True
    return None


class TestGEOFlippableFileYAMLReader(unittest.TestCase):
    """Test GEOFlippableFileYAMLReader."""

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch.object(yr.FileYAMLReader, "_load_dataset_with_area")
    def test_load_dataset_with_area_for_single_areas(self, ldwa):
        """Test _load_dataset_with_area() for single area definitions."""
        from pyresample.geometry import AreaDefinition

        from satpy.readers.core.yaml_reader import GEOFlippableFileYAMLReader

        reader = GEOFlippableFileYAMLReader()

        dsid = MagicMock()
        coords = MagicMock()

        # create a dummy upright xarray
        original_area_extent = (-1500, -1000, 1500, 1000)
        original_array = np.arange(6).reshape((2, 3))

        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            {"proj": "geos",
             "h": 35785831,
             "type": "crs"},
            3,
            2,
            original_area_extent,
        )

        dummy_ds_xr = xr.DataArray(original_array,
                                   coords={"y": np.arange(2),
                                           "x": np.arange(3),
                                           "time": ("y", np.arange(2))},
                                   attrs={"area": area_def},
                                   dims=("y", "x"))
        # assign the dummy xr as return for the super _load_dataset_with_area method
        ldwa.return_value = dummy_ds_xr

        # check no input, nothing should change
        res = reader._load_dataset_with_area(dsid, coords)
        np.testing.assert_equal(res.values, original_array)
        np.testing.assert_equal(res.attrs["area"].area_extent, original_area_extent)
        np.testing.assert_equal(res.coords["y"], np.arange(2))
        np.testing.assert_equal(res.coords["x"], np.arange(3))
        np.testing.assert_equal(res.coords["time"], np.arange(2))

        # check wrong input
        with pytest.raises(ValueError, match="Target orientation for Dataset unknown_name not recognized.*"):
            _ = reader._load_dataset_with_area(dsid, coords, "wronginput")

        # check native orientation, nothing should change
        res = reader._load_dataset_with_area(dsid, coords, "native")
        np.testing.assert_equal(res.values, original_array)
        np.testing.assert_equal(res.attrs["area"].area_extent, original_area_extent)
        np.testing.assert_equal(res.coords["y"], np.arange(2))
        np.testing.assert_equal(res.coords["x"], np.arange(3))
        np.testing.assert_equal(res.coords["time"], np.arange(2))

        # check upright orientation, nothing should change since area is already upright
        res = reader._load_dataset_with_area(dsid, coords, "NE")
        np.testing.assert_equal(res.values, original_array)
        np.testing.assert_equal(res.attrs["area"].area_extent, original_area_extent)
        np.testing.assert_equal(res.coords["y"], np.arange(2))
        np.testing.assert_equal(res.coords["x"], np.arange(3))
        np.testing.assert_equal(res.coords["time"], np.arange(2))

        # check that left-right image is flipped correctly
        dummy_ds_xr.attrs["area"] = area_def.copy(area_extent=(1500, -1000, -1500, 1000))
        ldwa.return_value = dummy_ds_xr.copy()
        res = reader._load_dataset_with_area(dsid, coords, "NE")
        np.testing.assert_equal(res.values, np.fliplr(original_array))
        np.testing.assert_equal(res.attrs["area"].area_extent, original_area_extent)
        np.testing.assert_equal(res.coords["y"], np.arange(2))
        np.testing.assert_equal(res.coords["x"], np.flip(np.arange(3)))
        np.testing.assert_equal(res.coords["time"], np.arange(2))

        # check that upside down image is flipped correctly
        dummy_ds_xr.attrs["area"] = area_def.copy(area_extent=(-1500, 1000, 1500, -1000))
        ldwa.return_value = dummy_ds_xr.copy()
        res = reader._load_dataset_with_area(dsid, coords, "NE")
        np.testing.assert_equal(res.values, np.flipud(original_array))
        np.testing.assert_equal(res.attrs["area"].area_extent, original_area_extent)
        np.testing.assert_equal(res.coords["y"], np.flip(np.arange(2)))
        np.testing.assert_equal(res.coords["x"], np.arange(3))
        np.testing.assert_equal(res.coords["time"], np.flip(np.arange(2)))

        # check different projection than geos, nothing should be changed
        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            {"proj": "lcc",
             "lat_1": 25.0,
             "type": "crs"},
            3,
            2,
            original_area_extent,
        )

        dummy_ds_xr = xr.DataArray(original_array,
                                   dims=("y", "x"),
                                   attrs={"area": area_def})
        ldwa.return_value = dummy_ds_xr
        res = reader._load_dataset_with_area(dsid, coords, "NE")
        np.testing.assert_equal(res.values, original_array)
        np.testing.assert_equal(res.attrs["area"].area_extent, original_area_extent)

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch.object(yr.FileYAMLReader, "_load_dataset_with_area")
    def test_load_dataset_with_area_for_stacked_areas(self, ldwa):
        """Test _load_dataset_with_area() for stacked area definitions."""
        from pyresample.geometry import AreaDefinition, StackedAreaDefinition

        from satpy.readers.core.yaml_reader import GEOFlippableFileYAMLReader

        reader = GEOFlippableFileYAMLReader()

        dsid = MagicMock()
        coords = MagicMock()

        # create a dummy upright xarray
        original_area_extents = [(-1500, -1000, 1500, 1000), (3000, 5000, 7000, 8000)]
        original_array = np.arange(12).reshape((4, 3))

        area_def0 = AreaDefinition(
            "test",
            "test",
            "test",
            {"proj": "geos",
             "h": 35785831,
             "type": "crs"},
            3,
            2,
            original_area_extents[0],
        )
        area_def1 = area_def0.copy(area_extent=original_area_extents[1])

        dummy_ds_xr = xr.DataArray(original_array,
                                   dims=("y", "x"),
                                   coords={"y": np.arange(4),
                                           "x": np.arange(3),
                                           "time": ("y", np.arange(4))},
                                   attrs={"area": StackedAreaDefinition(area_def0, area_def1)})

        # check that left-right image is flipped correctly
        dummy_ds_xr.attrs["area"].defs[0] = area_def0.copy(area_extent=(1500, -1000, -1500, 1000))
        dummy_ds_xr.attrs["area"].defs[1] = area_def1.copy(area_extent=(7000, 5000, 3000, 8000))
        ldwa.return_value = dummy_ds_xr.copy()
        res = reader._load_dataset_with_area(dsid, coords, "NE")
        np.testing.assert_equal(res.values, np.fliplr(original_array))
        np.testing.assert_equal(res.attrs["area"].defs[0].area_extent, original_area_extents[0])
        np.testing.assert_equal(res.attrs["area"].defs[1].area_extent, original_area_extents[1])
        np.testing.assert_equal(res.coords["y"], np.arange(4))
        np.testing.assert_equal(res.coords["x"], np.flip(np.arange(3)))
        np.testing.assert_equal(res.coords["time"], np.arange(4))

        # check that upside down image is flipped correctly
        dummy_ds_xr.attrs["area"].defs[0] = area_def0.copy(area_extent=(-1500, 1000, 1500, -1000))
        dummy_ds_xr.attrs["area"].defs[1] = area_def1.copy(area_extent=(3000, 8000, 7000, 5000))
        ldwa.return_value = dummy_ds_xr.copy()
        res = reader._load_dataset_with_area(dsid, coords, "NE")
        np.testing.assert_equal(res.values, np.flipud(original_array))
        # note that the order of the stacked areadefs is flipped here, as expected
        np.testing.assert_equal(res.attrs["area"].defs[1].area_extent, original_area_extents[0])
        np.testing.assert_equal(res.attrs["area"].defs[0].area_extent, original_area_extents[1])
        np.testing.assert_equal(res.coords["y"], np.flip(np.arange(4)))
        np.testing.assert_equal(res.coords["x"], np.arange(3))
        np.testing.assert_equal(res.coords["time"], np.flip(np.arange(4)))

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch.object(yr.FileYAMLReader, "_load_dataset_with_area")
    def test_load_dataset_with_area_for_swath_def_data(self, ldwa):
        """Test _load_dataset_with_area() for swath definition data."""
        from pyresample.geometry import SwathDefinition

        from satpy.readers.core.yaml_reader import GEOFlippableFileYAMLReader

        reader = GEOFlippableFileYAMLReader()

        dsid = MagicMock()
        coords = MagicMock()

        # create a dummy upright xarray
        original_array = np.ones(3)
        dim = np.arange(3)
        lats = np.arange(3)
        lons = np.arange(3)

        swath_def = SwathDefinition(lons, lats)
        dummy_ds_xr = xr.DataArray(original_array,
                                   coords={"y": dim},
                                   attrs={"area": swath_def},
                                   dims=("y",))

        # assign the dummy xr as return for the super _load_dataset_with_area method
        ldwa.return_value = dummy_ds_xr

        # returned dataset should be unchanged since datasets with a swath definition are not flippable
        res = reader._load_dataset_with_area(dsid, coords, "NE")
        np.testing.assert_equal(res.values, original_array)

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch.object(yr.FileYAMLReader, "_load_dataset_with_area")
    def test_load_dataset_with_area_for_data_without_area(self, ldwa):
        """Test _load_dataset_with_area() for data wihtout area information."""
        from satpy.readers.core.yaml_reader import GEOFlippableFileYAMLReader

        reader = GEOFlippableFileYAMLReader()

        dsid = MagicMock()
        coords = MagicMock()

        # create a dummy upright xarray
        original_array = np.ones(3)
        dim = np.arange(3)

        dummy_ds_xr = xr.DataArray(original_array,
                                   coords={"y": dim},
                                   attrs={},
                                   dims=("y",))

        # assign the dummy xr as return for the super _load_dataset_with_area method
        ldwa.return_value = dummy_ds_xr

        # returned dataset should be unchanged since datasets without area information are not flippable
        res = reader._load_dataset_with_area(dsid, coords, "NE")
        np.testing.assert_equal(res.values, original_array)


def _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, chk_pos_info):
    seg_area = MagicMock()
    seg_area.crs = "some_crs"
    seg_area.area_extent = aex
    seg_area.shape = ashape
    get_area_def = MagicMock()
    get_area_def.return_value = seg_area

    get_segment_position_info = MagicMock()
    get_segment_position_info.return_value = chk_pos_info

    filetype_info = {"expected_segments": expected_segments,
                     "file_type": "filetype1"}
    filename_info = {"segment": segment}

    fh = _create_mocked_basic_fh()
    fh.filetype_info = filetype_info
    fh.filename_info = filename_info
    fh.get_area_def = get_area_def
    fh.get_segment_position_info = get_segment_position_info

    return fh, seg_area


def _create_mocked_basic_fh():
    fake_fh = MagicMock()
    fake_fh.filename_info = {}
    fake_fh.filetype_info = {}
    return fake_fh

class TestGEOSegmentYAMLReader(unittest.TestCase):
    """Test GEOSegmentYAMLReader."""

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch.object(yr.FileYAMLReader, "create_filehandlers")
    def test_get_expected_segments(self, cfh):
        """Test that expected segments can come from the filename."""
        from satpy.readers.core.yaml_reader import GEOSegmentYAMLReader
        reader = GEOSegmentYAMLReader()

        fake_fh = _create_mocked_basic_fh()
        cfh.return_value = {"ft1": [fake_fh]}

        # default (1)
        created_fhs = reader.create_filehandlers(["fake.nc"])
        es = created_fhs["ft1"][0].filetype_info["expected_segments"]
        assert es == 1

        # YAML defined for each file type
        fake_fh.filetype_info["expected_segments"] = 2
        created_fhs = reader.create_filehandlers(["fake.nc"])
        es = created_fhs["ft1"][0].filetype_info["expected_segments"]
        assert es == 2

        # defined both in the filename and the YAML metadata
        # YAML has priority
        fake_fh.filename_info = {"total_segments": 3}
        fake_fh.filetype_info = {"expected_segments": 2}
        created_fhs = reader.create_filehandlers(["fake.nc"])
        es = created_fhs["ft1"][0].filetype_info["expected_segments"]
        assert es == 2

        # defined in the filename
        fake_fh.filename_info = {"total_segments": 3}
        fake_fh.filetype_info = {}
        created_fhs = reader.create_filehandlers(["fake.nc"])
        es = created_fhs["ft1"][0].filetype_info["expected_segments"]
        assert es == 3

        # check correct FCI segment (aka chunk in the FCI world) number reading into segment
        fake_fh.filename_info = {"count_in_repeat_cycle": 5}
        created_fhs = reader.create_filehandlers(["fake.nc"])
        es = created_fhs["ft1"][0].filename_info["segment"]
        assert es == 5

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch.object(yr.FileYAMLReader, "create_filehandlers")
    def test_segments_sorting(self, cfh):
        """Test that segment filehandlers are sorted by segment number."""
        from satpy.readers.core.yaml_reader import GEOSegmentYAMLReader
        reader = GEOSegmentYAMLReader()

        # create filehandlers with different segment numbers
        fake_fh_1 = _create_mocked_basic_fh()
        fake_fh_1.filename_info["segment"] = 1
        fake_fh_2 = _create_mocked_basic_fh()
        fake_fh_2.filename_info["segment"] = 2
        fake_fh_3 = _create_mocked_basic_fh()
        fake_fh_3.filename_info["segment"] = 3

        # put the filehandlers in an unsorted order
        reader.file_handlers = {"ft1": [fake_fh_1, fake_fh_3, fake_fh_2]}

        # check that the created filehandlers are sorted by segment number
        reader.create_filehandlers(["fake.nc"])
        assert [fh.filename_info["segment"] for fh in reader.file_handlers["ft1"]] == [1, 2, 3]

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch("satpy.readers.core.yaml_reader.FileYAMLReader._load_dataset")
    @patch("satpy.readers.core.yaml_reader.xr")
    @patch("satpy.readers.core.yaml_reader._find_missing_segments")
    def test_load_dataset(self, mss, xr, parent_load_dataset):
        """Test _load_dataset()."""
        from satpy.readers.core.yaml_reader import GEOSegmentYAMLReader
        reader = GEOSegmentYAMLReader()

        # Projectable is None
        mss.return_value = [0, 0, 0, False, None]
        with pytest.raises(KeyError):
            res = reader._load_dataset(None, None, None)
        # Failure is True
        mss.return_value = [0, 0, 0, True, 0]
        with pytest.raises(KeyError):
            res = reader._load_dataset(None, None, None)

        # Setup input, and output of mocked functions
        counter = 9
        expected_segments = 8
        seg = MagicMock(dims=["y", "x"])
        slice_list = expected_segments * [seg, ]
        failure = False
        projectable = MagicMock()
        mss.return_value = (counter, expected_segments, slice_list,
                            failure, projectable)
        empty_segment = MagicMock()
        xr.full_like.return_value = empty_segment
        concat_slices = MagicMock()
        xr.concat.return_value = concat_slices
        dataid = MagicMock()
        ds_info = MagicMock()
        file_handlers = MagicMock()

        # No missing segments
        res = reader._load_dataset(dataid, ds_info, file_handlers)
        assert res.attrs is file_handlers[0].combine_info.return_value
        assert empty_segment not in slice_list

        # One missing segment in the middle
        slice_list[4] = None
        counter = 8
        mss.return_value = (counter, expected_segments, slice_list,
                            failure, projectable)
        res = reader._load_dataset(dataid, ds_info, file_handlers)
        assert slice_list[4] is empty_segment

        # The last segment is missing
        slice_list = expected_segments * [seg, ]
        slice_list[-1] = None
        counter = 8
        mss.return_value = (counter, expected_segments, slice_list,
                            failure, projectable)
        res = reader._load_dataset(dataid, ds_info, file_handlers)
        assert slice_list[-1] is empty_segment

        # The last two segments are missing
        slice_list = expected_segments * [seg, ]
        slice_list[-1] = None
        counter = 7
        mss.return_value = (counter, expected_segments, slice_list,
                            failure, projectable)
        res = reader._load_dataset(dataid, ds_info, file_handlers)
        assert slice_list[-1] is empty_segment
        assert slice_list[-2] is empty_segment

        # The first segment is missing
        slice_list = expected_segments * [seg, ]
        slice_list[0] = None
        counter = 9
        mss.return_value = (counter, expected_segments, slice_list,
                            failure, projectable)
        res = reader._load_dataset(dataid, ds_info, file_handlers)
        assert slice_list[0] is empty_segment

        # The first two segments are missing
        slice_list = expected_segments * [seg, ]
        slice_list[0] = None
        slice_list[1] = None
        counter = 9
        mss.return_value = (counter, expected_segments, slice_list,
                            failure, projectable)
        res = reader._load_dataset(dataid, ds_info, file_handlers)
        assert slice_list[0] is empty_segment
        assert slice_list[1] is empty_segment

        # Disable padding
        res = reader._load_dataset(dataid, ds_info, file_handlers,
                                   pad_data=False)
        parent_load_dataset.assert_called_once_with(dataid, ds_info,
                                                    file_handlers)

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch("satpy.readers.core.yaml_reader._load_area_def")
    @patch("satpy.readers.core.yaml_reader._stack_area_defs")
    @patch("satpy.readers.core.yaml_reader.GEOSegmentYAMLReader._pad_earlier_segments_area")
    @patch("satpy.readers.core.yaml_reader.GEOSegmentYAMLReader._pad_later_segments_area")
    def test_load_area_def(self, pesa, plsa, sad, parent_load_area_def):
        """Test _load_area_def()."""
        from satpy.readers.core.yaml_reader import GEOSegmentYAMLReader
        reader = GEOSegmentYAMLReader()

        dataid = MagicMock()
        file_handlers = MagicMock()
        reader._load_area_def(dataid, file_handlers)
        pesa.assert_called_once()
        plsa.assert_called_once()
        sad.assert_called_once()
        parent_load_area_def.assert_not_called()
        # Disable padding
        reader._load_area_def(dataid, file_handlers, pad_data=False)
        parent_load_area_def.assert_called_once_with(dataid, file_handlers)

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch("satpy.readers.core.yaml_reader.AreaDefinition")
    def test_pad_later_segments_area(self, AreaDefinition):
        """Test _pad_later_segments_area()."""
        from satpy.readers.core.yaml_reader import GEOSegmentYAMLReader
        reader = GEOSegmentYAMLReader()

        expected_segments = 2
        segment = 1
        aex = [0, 1000, 200, 500]
        ashape = [200, 500]
        fh_1, _ = _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, None)
        file_handlers = [fh_1]
        dataid = "dataid"
        res = reader._pad_later_segments_area(file_handlers, dataid)
        assert len(res) == 2
        seg2_extent = (0, 1500, 200, 1000)
        expected_call = ("fill", "fill", "fill", "some_crs", 500, 200,
                         seg2_extent)
        AreaDefinition.assert_called_once_with(*expected_call)

    @patch.object(yr.FileYAMLReader, "__init__", lambda x: None)
    @patch("satpy.readers.core.yaml_reader.AreaDefinition")
    def test_pad_earlier_segments_area(self, AreaDefinition):
        """Test _pad_earlier_segments_area()."""
        from satpy.readers.core.yaml_reader import GEOSegmentYAMLReader
        reader = GEOSegmentYAMLReader()

        expected_segments = 2
        segment = 2
        aex = [0, 1000, 200, 500]
        ashape = [200, 500]
        fh_2, seg2_area = _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, None)

        file_handlers = [fh_2]
        dataid = "dataid"
        area_defs = {2: seg2_area}
        res = reader._pad_earlier_segments_area(file_handlers, dataid, area_defs)
        assert len(res) == 2
        seg1_extent = (0, 500, 200, 0)
        expected_call = ("fill", "fill", "fill", "some_crs", 500, 200,
                         seg1_extent)
        AreaDefinition.assert_called_once_with(*expected_call)

    def test_find_missing_segments(self):
        """Test _find_missing_segments()."""
        from satpy.readers.core.yaml_reader import _find_missing_segments as fms

        # Dataset with only one segment
        filename_info = {"segment": 1}
        fh_seg1 = MagicMock(filename_info=filename_info)
        projectable = "projectable"
        get_dataset = MagicMock()
        get_dataset.return_value = projectable
        fh_seg1.get_dataset = get_dataset
        file_handlers = [fh_seg1]
        ds_info = {"file_type": []}
        dataid = "dataid"
        res = fms(file_handlers, ds_info, dataid)
        counter, expected_segments, slice_list, failure, proj = res
        assert counter == 2
        assert expected_segments == 1
        assert projectable in slice_list
        assert failure is False
        assert proj is projectable

        # Three expected segments, first and last missing
        filename_info = {"segment": 2}
        filetype_info = {"expected_segments": 3,
                         "file_type": "foo"}
        fh_seg2 = MagicMock(filename_info=filename_info,
                            filetype_info=filetype_info)
        projectable = "projectable"
        get_dataset = MagicMock()
        get_dataset.return_value = projectable
        fh_seg2.get_dataset = get_dataset
        file_handlers = [fh_seg2]
        ds_info = {"file_type": ["foo"]}
        dataid = "dataid"
        res = fms(file_handlers, ds_info, dataid)
        counter, expected_segments, slice_list, failure, proj = res
        assert counter == 3
        assert expected_segments == 3
        assert slice_list == [None, projectable, None]
        assert failure is False
        assert proj is projectable


@pytest.fixture
@patch.object(yr.GEOVariableSegmentYAMLReader, "__init__", lambda x: None)
def GVSYReader():
    """Get a fixture of the GEOVariableSegmentYAMLReader."""
    from satpy.readers.core.yaml_reader import GEOVariableSegmentYAMLReader
    reader = GEOVariableSegmentYAMLReader()
    reader.segment_infos = dict()
    reader.segment_heights = cache(reader._segment_heights)
    return reader


@pytest.fixture
def fake_geswh():
    """Get a fixture of the patched _get_empty_segment_with_height."""
    with patch("satpy.readers.core.yaml_reader._get_empty_segment_with_height") as geswh:
        yield geswh


@pytest.fixture
def fake_xr():
    """Get a fixture of the patched xarray."""
    with patch("satpy.readers.core.yaml_reader.xr") as xr:
        yield xr


@pytest.fixture
def fake_mss():
    """Get a fixture of the patched _find_missing_segments."""
    with patch("satpy.readers.core.yaml_reader._find_missing_segments") as mss:
        yield mss


@pytest.fixture
def fake_adef():
    """Get a fixture of the patched AreaDefinition."""
    with patch("satpy.readers.core.yaml_reader.AreaDefinition") as adef:
        yield adef


class TestGEOVariableSegmentYAMLReader:
    """Test GEOVariableSegmentYAMLReader."""

    def test_get_empty_segment(self, GVSYReader, fake_mss, fake_xr, fake_geswh):
        """Test execution of (overridden) get_empty_segment inside _load_dataset."""
        # Setup input, and output of mocked functions for first segment missing
        chk_pos_info = {
            "1km": {"start_position_row": 0,
                    "end_position_row": 0,
                    "segment_height": 0,
                    "grid_width": 11136},
            "2km": {"start_position_row": 140,
                    "end_position_row": None,
                    "segment_height": 278,
                    "grid_width": 5568}
        }
        expected_segments = 2
        segment = 2
        aex = [0, 1000, 200, 500]
        ashape = [278, 5568]
        fh_2, _ = _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, chk_pos_info)

        GVSYReader.file_handlers = {"filetype1": [fh_2]}

        counter = 2
        seg = MagicMock(dims=["y", "x"])
        slice_list = [None, seg]
        failure = False
        projectable = MagicMock()
        empty_segment = MagicMock()
        empty_segment.shape = [278, 5568]
        fake_xr.full_like.return_value = empty_segment
        dataid = MagicMock()
        ds_info = MagicMock()

        fake_mss.return_value = (counter, expected_segments, slice_list,
                                 failure, projectable)
        GVSYReader._load_dataset(dataid, ds_info, [fh_2])
        # the return of get_empty_segment
        fake_geswh.assert_called_once_with(empty_segment, 139, dim="y")

    def test_pad_earlier_segments_area(self, GVSYReader, fake_adef):
        """Test _pad_earlier_segments_area() for the variable segment case."""
        # setting to 0 or None values that shouldn't be relevant
        chk_pos_info = {
            "1km": {"start_position_row": 0,
                    "end_position_row": 0,
                    "segment_height": 0,
                    "grid_width": 11136},
            "2km": {"start_position_row": 140,
                    "end_position_row": None,
                    "segment_height": 278,
                    "grid_width": 5568}
        }
        expected_segments = 2
        segment = 2
        aex = [0, 1000, 200, 500]
        ashape = [278, 5568]
        fh_2, seg2_area = _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, chk_pos_info)

        GVSYReader.file_handlers = {"filetype1": [fh_2]}
        dataid = "dataid"
        area_defs = {2: seg2_area}
        res = GVSYReader._pad_earlier_segments_area([fh_2], dataid, area_defs)
        assert len(res) == 2

        # The later vertical segment (nr. 2) size is 278, which is exactly double the size
        # of the gap left by the missing first segment (139, as the second segment starts at line 140).
        # Therefore, the new vertical area extent for the first segment should be
        # half of the previous size (1000-500)/2=250.
        # The new area extent lower-left row is therefore 500-250=250
        seg1_extent = (0, 500, 200, 250)
        expected_call = ("fill", "fill", "fill", "some_crs", 5568, 139,
                         seg1_extent)
        fake_adef.assert_called_once_with(*expected_call)

    def test_pad_later_segments_area(self, GVSYReader, fake_adef):
        """Test _pad_later_segments_area() in the variable padding case."""
        chk_pos_info = {
            "1km": {"start_position_row": None,
                    "end_position_row": 11136 - 278,
                    "segment_height": 556,
                    "grid_width": 11136},
            "2km": {"start_position_row": 0,
                    "end_position_row": 0,
                    "segment_height": 0,
                    "grid_width": 5568}}

        expected_segments = 2
        segment = 1
        aex = [0, 1000, 200, 500]
        ashape = [556, 11136]
        fh_1, _ = _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, chk_pos_info)
        GVSYReader.file_handlers = {"filetype1": [fh_1]}
        dataid = "dataid"
        res = GVSYReader._pad_later_segments_area([fh_1], dataid)
        assert len(res) == 2

        # The previous segment size is 556, which is exactly double the size of the gap left
        # by the missing last segment (278, as the second-to-last segment ends at line 11136 - 278 )
        # therefore, the new vertical area extent should be half of the previous size (1000-500)/2=250.
        # The new area extent lower-left row is therefore 1000+250=1250
        seg2_extent = (0, 1250, 200, 1000)
        expected_call = ("fill", "fill", "fill", "some_crs", 11136, 278,
                         seg2_extent)
        fake_adef.assert_called_once_with(*expected_call)

    def test_pad_later_segments_area_for_multiple_segments_gap(self, GVSYReader, fake_adef):
        """Test _pad_later_segments_area() in the variable padding case for multiple gaps with multiple segments."""

        def side_effect_areadef(a, b, c, crs, width, height, aex):
            m = MagicMock()
            m.shape = [height, width]
            m.area_extent = aex
            m.crs = crs
            return m

        fake_adef.side_effect = side_effect_areadef

        chk_pos_info = {
            "1km": {"start_position_row": 11136 - 600 - 100 + 1,
                    "end_position_row": 11136 - 600,
                    "segment_height": 100,
                    "grid_width": 11136},
            "2km": {"start_position_row": 0,
                    "end_position_row": 0,
                    "segment_height": 0,
                    "grid_width": 5568}}
        expected_segments = 8
        segment = 1
        aex = [0, 1000, 200, 500]
        ashape = [100, 11136]
        fh_1, _ = _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, chk_pos_info)
        chk_pos_info = {
            "1km": {"start_position_row": 11136 - 300 - 100 + 1,
                    "end_position_row": 11136 - 300,
                    "segment_height": 100,
                    "grid_width": 11136},
            "2km": {"start_position_row": 0,
                    "end_position_row": 0,
                    "segment_height": 0,
                    "grid_width": 5568}}
        segment = 4
        fh_4, _ = _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, chk_pos_info)

        chk_pos_info = {
            "1km": {"start_position_row": 11136 - 100 + 1,
                    "end_position_row": None,
                    "segment_height": 100,
                    "grid_width": 11136},
            "2km": {"start_position_row": 0,
                    "end_position_row": 0,
                    "segment_height": 0,
                    "grid_width": 5568}}
        segment = 8
        fh_8, _ = _create_mocked_fh_and_areadef(aex, ashape, expected_segments, segment, chk_pos_info)

        GVSYReader.file_handlers = {"filetype1": [fh_1, fh_4, fh_8]}
        dataid = "dataid"
        res = GVSYReader._pad_later_segments_area([fh_1, fh_4, fh_8], dataid)
        assert len(res) == 8

        # Regarding the segment sizes:
        # First group of missing segments:
        # The end position row of the gap is the start row of the last available segment-1:11136-300-100+1-1=10736
        # The start position row of the gap is the end row fo the first available segment+1: 11136-600+1=10837
        # hence the gap is 10736-10537+1=200 px high
        # The 200px have to be split between two missing segments, the most equal way to do it is with
        # sizes 100: 100+100=200
        # Second group:
        # The end position row of the gap is the start row of the last segment -1: 11136-100+1-1=11036
        # The start position row of the gap is the end row fo the first segment +1: 11136-300+1=10837
        # hence the gap is 11036-10837+1=200 px high
        # The 200px have to be split between three missing segments, the most equal way to do it is with
        # sizes 66 and 67: 66+67+67=200

        # Regarding the heights:
        # First group:
        # The first segment has 100px height and 500 area extent height.
        # The first padded segment has 100px height -> 500*100/100=500 area extent height ->1000+500=1500
        # The second padded segment has 100px height -> 500*100/100=500 area extent height ->1500+500=2000
        # Second group:
        # The first segment has 100px height and 500 area extent height.
        # The first padded segment has 66px height -> 500*66/100=330 area extent height ->1000+330=1330
        # The second padded segment has 67px height -> 500*67/100=335 area extent height ->1330+335=1665
        # The first padded segment has 67px height -> 500*67/100=335 area extent height ->1665+335=2000
        assert fake_adef.call_count == 5
        expected_call1 = ("fill", "fill", "fill", "some_crs", 11136, 100,
                          (0, 1500.0, 200, 1000))
        expected_call2 = ("fill", "fill", "fill", "some_crs", 11136, 100,
                          (0, 2000.0, 200, 1500))
        expected_call3 = ("fill", "fill", "fill", "some_crs", 11136, 66,
                          (0, 1330.0, 200, 1000))
        expected_call4 = ("fill", "fill", "fill", "some_crs", 11136, 67,
                          (0, 1665.0, 200, 1330.0))
        expected_call5 = ("fill", "fill", "fill", "some_crs", 11136, 67,
                          (0, 2000.0, 200, 1665.0))

        fake_adef.side_effect = None
        fake_adef.assert_has_calls([call(*expected_call1),
                                    call(*expected_call2),
                                    call(*expected_call3),
                                    call(*expected_call4),
                                    call(*expected_call5)
                                    ])

    @pytest.mark.parametrize("new_height", [
        139,  # same size
        140,  # larger size
        2000,  # much larger size
        138, # smaller size
        1, # much smaller size
    ])
    def test_get_empty_segment_with_height(self, new_height):
        """Test _get_empty_segment_with_height() for different new heights."""
        from satpy.readers.core.yaml_reader import _get_empty_segment_with_height as geswh

        empty_segment = xr.DataArray(da.ones((139, 5568)), dims=["y", "x"])
        new_empty_segment = geswh(empty_segment, new_height, "y")
        assert new_empty_segment.shape == (new_height, 5568)
        assert (new_empty_segment == empty_segment[0,0]).all()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""GERB L2 HR HDF5 reader.

Introduction
------------

A reader for the Top of Atmosphere outgoing fluxes from the Geostationary Earth Radiation
Budget instrument aboard the Meteosat Second Generation satellites.

Reader Arguments
----------------

Some arguments can be provided to the reader to change its behaviour. These are
provided through the `Scene` instantiation, eg::

  scn = Scene(filenames=filenames, reader="gerb_l2_hr_h5", reader_kwargs={"area": "msg_seviri_fes_9km"})


"""

import datetime as dt
import logging

from satpy.area import get_area_def
from satpy.readers.core.hdf5 import HDF5FileHandler

LOG = logging.getLogger(__name__)


def gerb_get_dataset(ds, ds_info):
    """Load a GERB dataset in memory from a HDF5 file or HDF5FileHandler.

    The routine takes into account the quantisation factor and fill values.
    """
    ds_attrs = ds.attrs
    ds_fill = ds_info["fill_value"]
    fill_mask = ds != ds_fill
    if "Quantisation Factor" in ds_attrs and "Unit" in ds_attrs:
        ds = ds*ds_attrs["Quantisation Factor"]
    else:
        ds = ds*1.
    ds = ds.where(fill_mask)
    return ds


class GERB_HR_FileHandler(HDF5FileHandler):
    """File handler for GERB L2 High Resolution H5 files.

    **Overriding the ``area``**

    By default, the GERB HR reader looks at the attribute "Geolocation/attr/Nominal
    Satellite Longitude (degrees)" for selecting the area definition. This data is not
    available for the GERB-like products however, so you may need to override it with the
    ``area`` argument::

        scene = satpy.Scene(filenames, reader="gerb_l2_hr_h5",
                                       area="msg_seviri_fes_9km")

    """

    def __init__(self, filename, filename_info, filetype_info, area=None):
        """Initialize the reader."""
        super(GERB_HR_FileHandler, self).__init__(filename, filename_info, filetype_info)
        self.user_area = None
        if area is not None:
            self.user_area = area

    @property
    def end_time(self):
        """Get end time."""
        return self.start_time + dt.timedelta(minutes=15)

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["sensing_time"]

    def get_dataset(self, ds_id, ds_info):
        """Read a HDF5 file into an xarray DataArray."""
        ds_name = ds_id["name"]
        if ds_name not in ["Solar Flux", "Thermal Flux", "Solar Radiance", "Thermal Radiance"]:
            raise KeyError(f"{ds_name} is an unknown dataset for this reader.")

        ds = gerb_get_dataset(self[f"Radiometry/{ds_name}"], ds_info)

        ds.attrs.update({"start_time": self.start_time, "data_time": self.start_time, "end_time": self.end_time})

        return ds

    def get_area_def(self, dsid):
        """Area definition for the GERB product."""
        if self.user_area is not None:
            return get_area_def(self.user_area)

        ssp_lon = self.file_content["Geolocation/attr/Nominal Satellite Longitude (degrees)"]

        if abs(ssp_lon) < 1e-6:
            return get_area_def("msg_seviri_fes_9km")
        elif abs(ssp_lon - 9.5) < 1e-6:
            return get_area_def("msg_seviri_fes_9km")
        elif abs(ssp_lon - 45.5) < 1e-6:
            return get_area_def("msg_seviri_iodc_9km")
        else:
            raise ValueError(f"There is no matching grid for SSP longitude {self.ssp_lon}")



# Copyright (c) 2010-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for saving-related functionality in scene.py."""

import datetime as dt
import os
from unittest import mock

import pytest
import xarray as xr
from dask import array as da

from satpy import Scene
from satpy.tests.utils import make_cid, spy_decorator

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path

@pytest.fixture(scope="module")
def fake_area():
    """Create a fake area for test data."""
    from pyresample.geometry import AreaDefinition

    return AreaDefinition(
        "test",
        "test",
        "test",
        "EPSG:4326",
        200,
        100,
        (-2000, -1000, 2000, 1000),
    )


class TestSceneSaving:
    """Test the Scene's saving method."""

    def test_save_datasets_default(self, tmp_path, fake_area):
        """Save a dataset using 'save_datasets'."""
        ds1 = xr.DataArray(
            da.arange(100 * 200).reshape((100, 200)).rechunk(50),
            dims=("y", "x"),
            attrs={"name": "test",
                   "start_time": dt.datetime(2018, 1, 1, 0, 0, 0),
                   "area": fake_area,
            }
        )
        scn = Scene()
        scn["test"] = ds1
        scn.save_datasets(base_dir=tmp_path)
        assert os.path.isfile(os.path.join(tmp_path, "test_20180101_000000.tif"))

    def test_save_datasets_by_ext(self, tmp_path):
        """Save a dataset using 'save_datasets' with 'filename'."""
        ds1 = xr.DataArray(
            da.arange(100 * 200).reshape((100, 200)).rechunk(50),
            dims=("y", "x"),
            attrs={"name": "test",
                   "start_time": dt.datetime(2018, 1, 1, 0, 0, 0),
                   "area": fake_area,
            }
        )
        scn = Scene()
        scn["test"] = ds1

        from satpy.writers.simple_image import PillowWriter
        save_image_mock = spy_decorator(PillowWriter.save_image)
        with mock.patch.object(PillowWriter, "save_image", save_image_mock):
            scn.save_datasets(base_dir=tmp_path, filename="{name}.png")
        save_image_mock.mock.assert_called_once()
        assert os.path.isfile(os.path.join(tmp_path, "test.png"))

    def test_save_datasets_bad_writer(self, tmp_path):
        """Save a dataset using 'save_datasets' and a bad writer."""
        ds1 = xr.DataArray(
            da.zeros((100, 200), chunks=50),
            dims=("y", "x"),
            attrs={"name": "test",
                   "start_time": dt.datetime.now(dt.timezone.utc)}
        )
        scn = Scene()
        scn["test"] = ds1
        pytest.raises(ValueError,
                      scn.save_datasets,
                      writer="_bad_writer_",
                      base_dir=tmp_path,
                      match="Unknown writer '_bad_writer_'")

    def test_save_datasets_missing_wishlist(self, tmp_path):
        """Calling 'save_datasets' with no valid datasets."""
        scn = Scene()
        scn._wishlist.add(make_cid(name="true_color"))
        pytest.raises(RuntimeError,
                      scn.save_datasets,
                      writer="geotiff",
                      base_dir=tmp_path)
        pytest.raises(KeyError,
                      scn.save_datasets,
                      datasets=["no_exist"])

    def test_save_dataset_default(self, tmp_path, fake_area):
        """Save a dataset using 'save_dataset'."""
        ds1 = xr.DataArray(
            da.arange(100 * 200).reshape((100, 200)).rechunk(50),
            dims=("y", "x"),
            attrs={"name": "test",
                   "start_time": dt.datetime(2018, 1, 1, 0, 0, 0),
                   "area": fake_area,
            }
        )
        scn = Scene()
        scn["test"] = ds1
        scn.save_dataset("test", base_dir=tmp_path)
        assert os.path.isfile(os.path.join(tmp_path, "test_20180101_000000.tif"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Fetch avhrr calibration coefficients."""
import datetime as dt
import os.path
import sys

import h5py
import urllib2

BASE_URL = "http://www.star.nesdis.noaa.gov/smcd/spb/fwu/homepage/" + \
           "AVHRR/Op_Cal_AVHRR/"

URLS = {
    "Metop-B":
    {"ch1": BASE_URL + "Metop1_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "Metop1_AVHRR_Libya_ch2.txt",
     "ch3a": BASE_URL + "Metop1_AVHRR_Libya_ch3a.txt"},
    "Metop-A":
    {"ch1": BASE_URL + "Metop2_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "Metop2_AVHRR_Libya_ch2.txt",
     "ch3a": BASE_URL + "Metop2_AVHRR_Libya_ch3a.txt"},
    "NOAA-16":
    {"ch1": BASE_URL + "N16_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "N16_AVHRR_Libya_ch2.txt"},
    "NOAA-17":
    {"ch1": BASE_URL + "N17_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "N17_AVHRR_Libya_ch2.txt",
     "ch3a": BASE_URL + "N17_AVHRR_Libya_ch3a.txt"},
    "NOAA-18":
    {"ch1": BASE_URL + "N18_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "N18_AVHRR_Libya_ch2.txt"},
    "NOAA-19":
    {"ch1": BASE_URL + "N19_AVHRR_Libya_ch1.txt",
     "ch2": BASE_URL + "N19_AVHRR_Libya_ch2.txt"}
}


def get_page(url):
    """Retrieve the given page."""
    return urllib2.urlopen(url).read()  # nosec


def get_coeffs(page):
    """Parse coefficients from the page."""
    coeffs = {}
    coeffs["datetime"] = []
    coeffs["slope1"] = []
    coeffs["intercept1"] = []
    coeffs["slope2"] = []
    coeffs["intercept2"] = []

    slope1_idx, intercept1_idx, slope2_idx, intercept2_idx = \
        None, None, None, None

    date_idx = 0
    for row in page.lower().split("\n"):
        row = row.split()
        if len(row) == 0:
            continue
        if row[0] == "update":
            # Get the column indices from the header line
            slope1_idx = row.index("slope_lo")
            intercept1_idx = row.index("int_lo")
            slope2_idx = row.index("slope_hi")
            intercept2_idx = row.index("int_hi")
            continue

        if slope1_idx is None:
            continue

        # In some cases the fields are connected, skip those rows
        if max([slope1_idx, intercept1_idx,
                slope2_idx, intercept2_idx]) >= len(row):
            continue

        try:
            dat = dt.datetime.strptime(row[date_idx], "%m/%d/%Y")
        except ValueError:
            continue

        coeffs["datetime"].append([dat.year, dat.month, dat.day])
        coeffs["slope1"].append(float(row[slope1_idx]))
        coeffs["intercept1"].append(float(row[intercept1_idx]))
        coeffs["slope2"].append(float(row[slope2_idx]))
        coeffs["intercept2"].append(float(row[intercept2_idx]))

    return coeffs


def get_all_coeffs():
    """Get all available calibration coefficients for the satellites."""
    coeffs = {}

    for platform in URLS:
        if platform not in coeffs:
            coeffs[platform] = {}
        for chan in URLS[platform].keys():
            url = URLS[platform][chan]
            print(url)
            page = get_page(url)
            coeffs[platform][chan] = get_coeffs(page)

    return coeffs


def save_coeffs(coeffs, out_dir=""):
    """Save calibration coefficients to HDF5 files."""
    for platform in coeffs.keys():
        fname = os.path.join(out_dir, "%s_calibration_data.h5" % platform)
        fid = h5py.File(fname, "w")

        for chan in coeffs[platform].keys():
            fid.create_group(chan)
            fid[chan]["datetime"] = coeffs[platform][chan]["datetime"]
            fid[chan]["slope1"] = coeffs[platform][chan]["slope1"]
            fid[chan]["intercept1"] = coeffs[platform][chan]["intercept1"]
            fid[chan]["slope2"] = coeffs[platform][chan]["slope2"]
            fid[chan]["intercept2"] = coeffs[platform][chan]["intercept2"]

        fid.close()
        print("Calibration coefficients saved for %s" % platform)


def main():
    """Create calibration coefficient files for AVHRR."""
    out_dir = sys.argv[1]
    coeffs = get_all_coeffs()
    save_coeffs(coeffs, out_dir=out_dir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Enhancement functions specific to the ABI sensor."""

from satpy.enhancements.wrappers import exclude_alpha, using_map_blocks


def cimss_true_color_contrast(img, **kwargs):
    """Scale data based on CIMSS True Color recipe for AWIPS."""
    _cimss_true_color_contrast(img.data)


@exclude_alpha
@using_map_blocks
def _cimss_true_color_contrast(img_data):
    """Perform per-chunk enhancement.

    Code ported from Kaba Bah's AWIPS python plugin for creating the
    CIMSS Natural (True) Color image in AWIPS. AWIPS provides that python
    code the image data on a 0-255 scale. Satpy gives this function the
    data on a 0-1.0 scale (assuming linear stretching and sqrt
    enhancements have already been applied).

    """
    max_value = 1.0
    acont = (255.0 / 10.0) / 255.0
    amax = (255.0 + 4.0) / 255.0
    amid = 1.0 / 2.0
    afact = (amax * (acont + max_value) / (max_value * (amax - acont)))
    aband = (afact * (img_data - amid) + amid)
    aband[aband <= 10 / 255.0] = 0
    aband[aband >= 1.0] = 1.0
    return aband



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the CF Area."""
import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample import AreaDefinition, SwathDefinition

from satpy.cf.area import _add_grid_mapping, area2cf


@pytest.fixture
def input_data_arr() -> xr.DataArray:
    """Create a data array."""
    return xr.DataArray(
        data=[[1, 2], [3, 4]],
        dims=("y", "x"),
        coords={"y": [1, 2], "x": [3, 4]},
        attrs={"name": "var1"},
    )


class TestCFArea:
    """Test case for CF Area."""

    @pytest.mark.parametrize("include_lonlats", [False, True])
    def test_area2cf_geos_area_nolonlats(self, input_data_arr, include_lonlats):
        """Test the conversion of an area to CF standards."""
        geos = AreaDefinition(
            area_id="geos",
            description="geos",
            proj_id="geos",
            projection={"proj": "geos", "h": 35785831., "a": 6378169., "b": 6356583.8},
            width=2, height=2,
            area_extent=[-1, -1, 1, 1])
        input_data_arr.attrs["area"] = geos

        res = area2cf(input_data_arr, include_lonlats=include_lonlats)
        assert len(res) == 2
        assert res[0].size == 1  # grid mapping variable
        assert res[0].name == res[1].attrs["grid_mapping"]
        if include_lonlats:
            assert "longitude" in res[1].coords
            assert "latitude" in res[1].coords

    def test_area2cf_swath(self, input_data_arr):
        """Test area2cf for swath definitions."""
        swath = SwathDefinition(lons=[[1, 1], [2, 2]], lats=[[1, 2], [1, 2]])
        input_data_arr.attrs["area"] = swath

        res = area2cf(input_data_arr, include_lonlats=False)
        assert len(res) == 1
        assert "longitude" in res[0].coords
        assert "latitude" in res[0].coords
        assert "grid_mapping" not in res[0].attrs

    def test_add_grid_mapping_cf_repr(self, input_data_arr):
        """Test the conversion from pyresample area object to CF grid mapping.

        Projection has a corresponding CF representation (e.g. geos).

        """
        a = 6378169.
        b = 6356583.8
        h = 35785831.
        geos = AreaDefinition(
            area_id="geos",
            description="geos",
            proj_id="geos",
            projection={"proj": "geos", "h": h, "a": a, "b": b,
                        "lat_0": 0, "lon_0": 0},
            width=2, height=2,
            area_extent=[-1, -1, 1, 1])
        geos_expected = xr.DataArray(data=0,
                                     attrs={"perspective_point_height": h,
                                            "latitude_of_projection_origin": 0,
                                            "longitude_of_projection_origin": 0,
                                            "grid_mapping_name": "geostationary",
                                            "semi_major_axis": a,
                                            "semi_minor_axis": b,
                                            # 'sweep_angle_axis': None,
                                            })

        input_data_arr.attrs["area"] = geos
        new_ds, grid_mapping = _add_grid_mapping(input_data_arr)
        if "sweep_angle_axis" in grid_mapping.attrs:
            # older versions of pyproj might not include this
            assert grid_mapping.attrs["sweep_angle_axis"] == "y"

        assert new_ds.attrs["grid_mapping"] == "geos"
        _gm_matches(grid_mapping, geos_expected)
        # should not have been modified
        assert "grid_mapping" not in input_data_arr.attrs

    def test_add_grid_mapping_no_cf_repr(self, input_data_arr):
        """Test the conversion from pyresample area object to CF grid mapping.

        Projection does not have a corresponding CF representation (e.g. COSMO).

        """
        cosmo7 = AreaDefinition(
            area_id="cosmo7",
            description="cosmo7",
            proj_id="cosmo7",
            projection={"proj": "ob_tran", "ellps": "WGS84", "lat_0": 46, "lon_0": 4.535,
                        "o_proj": "stere", "o_lat_p": 90, "o_lon_p": -5.465},
            width=597, height=510,
            area_extent=[-1812933, -1003565, 814056, 1243448]
        )
        input_data_arr.attrs["area"] = cosmo7

        new_ds, grid_mapping = _add_grid_mapping(input_data_arr)
        assert "crs_wkt" in grid_mapping.attrs
        wkt = grid_mapping.attrs["crs_wkt"]
        assert 'ELLIPSOID["WGS 84"' in wkt
        assert 'PARAMETER["lat_0",46' in wkt
        assert 'PARAMETER["lon_0",4.535' in wkt
        assert 'PARAMETER["o_lat_p",90' in wkt
        assert 'PARAMETER["o_lon_p",-5.465' in wkt
        assert new_ds.attrs["grid_mapping"] == "cosmo7"

    def test_add_grid_mapping_transverse_mercator(self, input_data_arr):
        """Test the conversion from pyresample area object to CF grid mapping.

        Projection is transverse mercator.

        """
        lat_0 = 36.5
        lon_0 = 15.0

        tmerc = AreaDefinition(
            area_id="tmerc",
            description="tmerc",
            proj_id="tmerc",
            projection={"proj": "tmerc", "ellps": "WGS84", "lat_0": 36.5, "lon_0": 15.0},
            width=2, height=2,
            area_extent=[-1, -1, 1, 1])

        tmerc_expected = xr.DataArray(data=0,
                                      attrs={"latitude_of_projection_origin": lat_0,
                                             "longitude_of_central_meridian": lon_0,
                                             "grid_mapping_name": "transverse_mercator",
                                             "reference_ellipsoid_name": "WGS 84",
                                             "false_easting": 0.,
                                             "false_northing": 0.,
                                             })

        input_data_arr.attrs["area"] = tmerc
        new_ds, grid_mapping = _add_grid_mapping(input_data_arr)
        assert new_ds.attrs["grid_mapping"] == "tmerc"
        _gm_matches(grid_mapping, tmerc_expected)

    def test_add_grid_mapping_cf_repr_no_ab(self, input_data_arr):
        """Test the conversion from pyresample area object to CF grid mapping.

        Projection has a corresponding CF representation but no explicit a/b.

        """
        h = 35785831.
        geos = AreaDefinition(
            area_id="geos",
            description="geos",
            proj_id="geos",
            projection={"proj": "geos", "h": h, "datum": "WGS84", "ellps": "GRS80",
                        "lat_0": 0, "lon_0": 0},
            width=2, height=2,
            area_extent=[-1, -1, 1, 1])
        geos_expected = xr.DataArray(data=0,
                                     attrs={"perspective_point_height": h,
                                            "latitude_of_projection_origin": 0,
                                            "longitude_of_projection_origin": 0,
                                            "grid_mapping_name": "geostationary",
                                            "reference_ellipsoid_name": "WGS 84",
                                            # 'semi_major_axis': 6378137.0,
                                            # 'semi_minor_axis': 6356752.314,
                                            # 'sweep_angle_axis': None,
                                            })

        input_data_arr.attrs["area"] = geos
        new_ds, grid_mapping = _add_grid_mapping(input_data_arr)

        assert new_ds.attrs["grid_mapping"] == "geos"
        _gm_matches(grid_mapping, geos_expected)

    def test_add_grid_mapping_oblique_mercator(self, input_data_arr):
        """Test the conversion from pyresample area object to CF grid mapping.

        Projection is oblique mercator.

        """
        area = AreaDefinition(
            area_id="omerc_otf",
            description="On-the-fly omerc area",
            proj_id="omerc",
            projection={"alpha": "9.02638777018478", "ellps": "WGS84", "gamma": "0", "k": "1",
                        "lat_0": "-0.256794486098476", "lonc": "13.7888658224205",
                        "proj": "omerc", "units": "m"},
            width=2837,
            height=5940,
            area_extent=[-1460463.0893, 3455291.3877, 1538407.1158, 9615788.8787]
        )

        omerc_dict = {"azimuth_of_central_line": 9.02638777018478,
                      "false_easting": 0.,
                      "false_northing": 0.,
                      # 'gamma': 0,  # this is not CF compliant
                      "grid_mapping_name": "oblique_mercator",
                      "latitude_of_projection_origin": -0.256794486098476,
                      "longitude_of_projection_origin": 13.7888658224205,
                      # 'prime_meridian_name': "Greenwich",
                      "reference_ellipsoid_name": "WGS 84"}
        omerc_expected = xr.DataArray(data=0, attrs=omerc_dict)

        input_data_arr.attrs["area"] = area
        new_ds, grid_mapping = _add_grid_mapping(input_data_arr)

        assert new_ds.attrs["grid_mapping"] == "omerc_otf"
        _gm_matches(grid_mapping, omerc_expected)

    @pytest.mark.parametrize("dims", [("y", "x"), ("bands", "y", "x")])
    def test_add_lonlat_coords(self, dims):
        """Test the conversion from areas to lon/lat."""
        from satpy.cf.area import _add_lonlat_coords

        width, height = (2, 2) if len(dims) == 2 else (10, 10)
        area = AreaDefinition(
            "seviri",
            "Native SEVIRI grid",
            "geos",
            "+a=6378169.0 +h=35785831.0 +b=6356583.8 +lon_0=0 +proj=geos",
            width, height,
            [-5570248.686685662, -5567248.28340708, 5567248.28340708, 5570248.686685662]
        )
        lons_ref, lats_ref = area.get_lonlats()
        if len(dims) == 2:
            data_arr = xr.DataArray(data=[[1, 2], [3, 4]], dims=dims, attrs={"area": area})
        else:
            data_arr = xr.DataArray(
                data=da.from_array(np.arange(3 * 10 * 10).reshape(3, 10, 10), chunks=(1, 5, 5)),
                dims=("bands", "y", "x"),
                attrs={"area": area},
            )

        res = _add_lonlat_coords(data_arr)

        # original should be unmodified
        assert "longitude" not in data_arr.coords
        assert set(res.coords) == {"longitude", "latitude"}
        lat = res["latitude"]
        lon = res["longitude"]
        np.testing.assert_array_equal(lat.data, lats_ref)
        np.testing.assert_array_equal(lon.data, lons_ref)
        assert {"name": "latitude", "standard_name": "latitude", "units": "degrees_north"}.items() <= lat.attrs.items()
        assert {"name": "longitude", "standard_name": "longitude", "units": "degrees_east"}.items() <= lon.attrs.items()


def _gm_matches(gmapping, expected):
    """Assert that all keys in ``expected`` match the values in ``gmapping``."""
    for attr_key, attr_val in expected.attrs.items():
        test_val = gmapping.attrs[attr_key]
        if attr_val is None or isinstance(attr_val, str):
            assert test_val == attr_val
        else:
            np.testing.assert_almost_equal(test_val, attr_val, decimal=3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to MiRS product."""

import datetime
import importlib
import logging
import os
from collections import Counter

import dask.array as da
import numpy as np
import xarray as xr

from satpy.aux_download import retrieve
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()
LOG = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def get_resource_string(mod_part, file_part):
    """Read resource string."""
    ref = importlib.resources.files(mod_part).joinpath(file_part)
    return ref.read_bytes()


# 'Polo' variable in MiRS files use these values for H/V polarization
POLO_V = 2
POLO_H = 3

amsu = "amsu-mhs"
PLATFORMS = {"n18": "NOAA-18",
             "n19": "NOAA-19",
             "np": "NOAA-19",
             "n20": "NOAA-20",
             "n21": "NOAA-21",
             "n22": "NOAA-22",
             "n23": "NOAA-23",
             "m2": "MetOp-A",
             "m1": "MetOp-B",
             "m3": "MetOp-C",
             "ma2": "MetOp-A",
             "ma1": "MetOp-B",
             "ma3": "MetOp-C",
             "npp": "NPP",
             "f17": "DMSP-F17",
             "f18": "DMSP-F18",
             "gpm": "GPM",
             }
SENSOR = {"n18": amsu,
          "n19": amsu,
          "n20": "atms",
          "n21": "atms",
          "n22": "atms",
          "n23": "atms",
          "n24": "atms",
          "np": amsu,
          "m1": amsu,
          "m2": amsu,
          "m3": amsu,
          "ma1": amsu,
          "ma2": amsu,
          "ma3": amsu,
          "npp": "atms",
          "jpss": "atms",
          "f17": "ssmis",
          "f18": "ssmis",
          "gpm": "GPI",
          }


def read_atms_coeff_to_string(fn):
    """Read the coefficients into a string."""
    if os.path.isfile(fn):
        coeff_str = open(fn, "r").readlines()
    else:
        parts = fn.split(":")
        mod_part, file_part = parts if len(parts) == 2 else ("", parts[0])
        mod_part = mod_part or __package__  # self.__module__
        coeff_str = get_resource_string(mod_part, file_part).decode().split("\n")

    return coeff_str


def read_atms_limb_correction_coefficients(fn):
    """Read the limb correction files."""
    coeff_str = read_atms_coeff_to_string(fn)
    n_chn = 22
    n_fov = 96
    # make the string a generator
    coeff_lines = (line.strip() for line in coeff_str)

    all_coeffs = np.zeros((n_chn, n_fov, n_chn), dtype=np.float32)
    all_amean = np.zeros((n_chn, n_fov, n_chn), dtype=np.float32)
    all_dmean = np.zeros(n_chn, dtype=np.float32)
    all_nchx = np.zeros(n_chn, dtype=np.int32)
    all_nchanx = np.zeros((n_chn, n_chn), dtype=np.int32)
    all_nchanx[:] = 9999
    # There should be 22 sections
    for chan_idx in range(n_chn):
        # blank line at the start of each section
        _ = next(coeff_lines)
        # section header
        next_line = next(coeff_lines)

        _nx, nchx, dmean = [x.strip() for x in next_line.split(" ") if x]
        all_nchx[chan_idx] = nchx = int(nchx)
        all_dmean[chan_idx] = float(dmean)

        # coeff locations (indexes to put the future coefficients in)
        next_line = next(coeff_lines)
        locations = [int(x.strip()) for x in next_line.split(" ") if x]
        if len(locations) != nchx:
            raise RuntimeError
        for x in range(nchx):
            all_nchanx[chan_idx, x] = locations[x] - 1

        # Read 'nchx' coefficients for each of 96 FOV
        for fov_idx in range(n_fov):
            # chan_num, fov_num, *coefficients, error
            coeff_line_parts = [x.strip() for x in next(coeff_lines).split(" ") if x][2:]
            coeffs = [float(x) for x in coeff_line_parts[:nchx]]
            ameans = [float(x) for x in coeff_line_parts[nchx:-1]]
            # not used but nice to know the purpose of the last column.
            # _error_val = float(coeff_line_parts[-1])
            for x in range(nchx):
                all_coeffs[chan_idx, fov_idx, all_nchanx[chan_idx, x]] = coeffs[x]
                all_amean[all_nchanx[chan_idx, x], fov_idx, chan_idx] = ameans[x]

    return all_dmean, all_coeffs, all_amean, all_nchx, all_nchanx


def apply_atms_limb_correction(datasets, channel_idx, dmean,
                               coeffs, amean, nchx, nchanx):
    """Calculate the correction for each channel."""
    ds = datasets[channel_idx]

    fov_line_correct = []
    for fov_idx in range(ds.shape[1]):
        coeff_sum = np.zeros(ds.shape[0], dtype=ds.dtype)
        for k in range(nchx[channel_idx]):
            chn_repeat = nchanx[channel_idx, k]
            coef = coeffs[channel_idx, fov_idx, chn_repeat] * (
                   datasets[chn_repeat, :, fov_idx] -
                   amean[chn_repeat, fov_idx, channel_idx])
            coeff_sum = np.add(coef, coeff_sum)
        fov_line_correct.append(np.add(coeff_sum, dmean[channel_idx]))
    return np.stack(fov_line_correct, axis=1)


def get_coeff_by_sfc(coeff_fn, bt_data, idx):
    """Read coefficients for specific filename (land or sea)."""
    sfc_coeff = read_atms_limb_correction_coefficients(coeff_fn)
    # transpose bt_data for correction
    bt_data = bt_data.transpose("Channel", "y", "x")
    c_size = bt_data[idx, :, :].chunks
    correction = da.map_blocks(apply_atms_limb_correction,
                               bt_data, idx,
                               *sfc_coeff, chunks=c_size, meta=np.array((), dtype=bt_data.dtype))
    return correction


def limb_correct_atms_bt(bt_data, surf_type_mask, coeff_fns, ds_info):
    """Gather data needed for limb correction."""
    idx = ds_info["channel_index"]
    LOG.info("Starting ATMS Limb Correction...")

    sea_bt = get_coeff_by_sfc(coeff_fns["sea"], bt_data, idx)
    land_bt = get_coeff_by_sfc(coeff_fns["land"], bt_data, idx)

    LOG.info("Finishing limb correction")
    is_sea = (surf_type_mask == 0)
    new_data = np.where(is_sea, sea_bt, land_bt)

    bt_corrected = xr.DataArray(new_data, dims=("y", "x"),
                                attrs=ds_info)

    return bt_corrected


class MiRSL2ncHandler(BaseFileHandler):
    """MiRS handler for NetCDF4 files using xarray.

    The MiRS retrieval algorithm runs on multiple
    sensors.  For the ATMS sensors, a limb correction
    is applied by default.  In order to change that
    behavior, use the keyword argument ``limb_correction=False``::


        from satpy import Scene, find_files_and_readers

        filenames = find_files_and_readers(base_dir, reader="mirs")
        scene = Scene(filenames, reader_kwargs={'limb_correction': False})

    """

    def __init__(self, filename, filename_info, filetype_info,
                 limb_correction=True):
        """Init method."""
        super(MiRSL2ncHandler, self).__init__(filename,
                                              filename_info,
                                              filetype_info,
                                              )

        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  decode_coords=True,
                                  chunks={"Field_of_view": CHUNK_SIZE,
                                          "Scanline": CHUNK_SIZE})
        # y,x is used in satpy, bands rather than channel using in xrimage
        self.nc = self.nc.rename_dims({"Scanline": "y",
                                       "Field_of_view": "x"})
        self.nc = self.nc.rename({"Latitude": "latitude",
                                  "Longitude": "longitude"})

        self.platform_name = self._get_platform_name
        self.sensor = self._get_sensor
        self.limb_correction = limb_correction

    @property
    def platform_shortname(self):
        """Get platform shortname."""
        return self.filename_info["platform_shortname"]

    @property
    def _get_platform_name(self):
        """Get platform name."""
        try:
            res = PLATFORMS[self.filename_info["platform_shortname"].lower()]
        except KeyError:
            res = "mirs"
        return res.lower()

    @property
    def _get_sensor(self):
        """Get sensor."""
        try:
            res = SENSOR[self.filename_info["platform_shortname"].lower()]
        except KeyError:
            res = self.sensor_names
        return res

    @property
    def sensor_names(self):
        """Return standard sensor names for the file's data."""
        return list(set(SENSOR.values()))

    @property
    def start_time(self):
        """Get start time."""
        # old file format
        if self.filename_info.get("date", False):
            s_time = datetime.datetime.combine(
                self.force_date("date"),
                self.force_time("start_time")
            )
            self.filename_info["start_time"] = s_time
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get end time."""
        # old file format
        if self.filename_info.get("date", False):
            end_time = datetime.datetime.combine(
                self.force_date("date"),
                self.force_time("end_time")
            )
            self.filename_info["end_time"] = end_time
        return self.filename_info["end_time"]

    def force_date(self, key):
        """Force datetime.date for combine."""
        if isinstance(self.filename_info[key], datetime.datetime):
            return self.filename_info[key].date()
        return self.filename_info[key]

    def force_time(self, key):
        """Force datetime.time for combine."""
        if isinstance(self.filename_info.get(key), datetime.datetime):
            return self.filename_info.get(key).time()
        return self.filename_info.get(key)

    @property
    def _get_coeff_filenames(self):
        """Retrieve necessary files for coefficients if needed."""
        coeff_fn = {"sea": None, "land": None}
        if self.platform_name.startswith("noaa"):
            suffix = self.platform_name[-2:]
            coeff_fn["land"] = retrieve(f"readers/limbcoef_atmsland_noaa{suffix}.txt")
            coeff_fn["sea"] = retrieve(f"readers/limbcoef_atmssea_noaa{suffix}.txt")
        if self.platform_name == "npp":
            coeff_fn["land"] = retrieve("readers/limbcoef_atmsland_snpp.txt")
            coeff_fn["sea"] = retrieve("readers/limbcoef_atmssea_snpp.txt")

        return coeff_fn

    def update_metadata(self, ds_info):
        """Get metadata."""
        metadata = {}
        metadata.update(ds_info)
        metadata.update({
            "sensor": self.sensor,
            "platform_name": self.platform_name,
            "start_time": self.start_time,
            "end_time": self.end_time,
        })
        return metadata

    @staticmethod
    def _nan_for_dtype(data_arr_dtype):
        # don't force the conversion from 32-bit float to 64-bit float
        # if we don't have to
        if data_arr_dtype.type == np.float32:
            return np.float32(np.nan)
        if np.issubdtype(data_arr_dtype, np.timedelta64):
            return np.timedelta64("NaT")
        if np.issubdtype(data_arr_dtype, np.datetime64):
            return np.datetime64("NaT")
        return np.float32(np.nan)

    @staticmethod
    def _scale_data(data_arr, scale_factor, add_offset):
        """Scale data, if needed."""
        scaling_needed = not (scale_factor == 1 and add_offset == 0)
        if scaling_needed:
            data_arr = data_arr * np.float32(scale_factor) + np.float32(add_offset)
        return data_arr

    def _fill_data(self, data_arr, fill_value, scale_factor, add_offset):
        """Fill missing data with NaN."""
        if fill_value is not None:
            # NOTE: Sfc_type and other category products are not detected or handled properly
            #   and will be converted from integers to 32-bit floats in this step
            fill_value = self._scale_data(fill_value, scale_factor, add_offset)
            fill_out = self._nan_for_dtype(data_arr.dtype)
            data_arr = data_arr.where(data_arr != fill_value, fill_out)
        return data_arr

    def _apply_valid_range(self, data_arr, valid_range, scale_factor, add_offset):
        """Get and apply valid_range."""
        if valid_range is not None:
            valid_min, valid_max = valid_range
            valid_min = self._scale_data(valid_min, scale_factor, add_offset)
            valid_max = self._scale_data(valid_max, scale_factor, add_offset)

            if valid_min is not None and valid_max is not None:
                data_arr = data_arr.where((data_arr >= valid_min) &
                                          (data_arr <= valid_max))
        return data_arr

    def apply_attributes(self, data, ds_info):
        """Combine attributes from file and yaml and apply.

        File attributes should take precedence over yaml if both are present

        """
        try:
            global_attr_fill = self.nc.attrs["missing_value"]
        except AttributeError:
            global_attr_fill = 1.0

        # let file metadata take precedence over ds_info from yaml,
        # but if yaml has more to offer, include it here, but fix
        # units.
        ds_info.update(data.attrs)

        # special cases
        if ds_info["name"] in ["latitude", "longitude"]:
            ds_info["standard_name"] = ds_info.get("standard_name",
                                                   ds_info["name"])

        # try to assign appropriate units (if "Kelvin" covert to K)
        units_convert = {"Kelvin": "K"}
        data_unit = ds_info.get("units", None)
        ds_info["units"] = units_convert.get(data_unit, data_unit)

        scale = ds_info.pop("scale_factor", 1.0)
        offset = ds_info.pop("add_offset", 0.)
        fill_value = ds_info.pop("_FillValue", global_attr_fill)
        valid_range = ds_info.pop("valid_range", None)

        data = self._scale_data(data, scale, offset)
        data = self._fill_data(data, fill_value, scale, offset)
        data = self._apply_valid_range(data, valid_range, scale, offset)

        data.attrs = ds_info

        return data, ds_info

    def get_dataset(self, ds_id, ds_info):
        """Get datasets."""
        if "dependencies" in ds_info.keys():
            idx = ds_info["channel_index"]
            data = self["BT"]
            data = data.rename(new_name_or_name_dict=ds_info["name"])
            data, ds_info = self.apply_attributes(data, ds_info)

            if self.sensor.lower() == "atms" and self.limb_correction:
                sfc_type_mask = self["Sfc_type"]
                data = limb_correct_atms_bt(data, sfc_type_mask,
                                            self._get_coeff_filenames,
                                            ds_info)

                self.nc = self.nc.merge(data)
            else:
                LOG.info("No Limb Correction applied.")
                data = data[:, :, idx]
        else:
            data = self[ds_id["name"]]
            data, ds_info = self.apply_attributes(data, ds_info)

        data.attrs = self.update_metadata(ds_info)

        return data

    def available_datasets(self, configured_datasets=None):
        """Dynamically discover what variables can be loaded from this file.

        See :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for more information.

        """
        handled_vars = set()
        for is_avail, ds_info in (configured_datasets or []):
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info
                continue

            yaml_info = {}
            if self.file_type_matches(ds_info["file_type"]):
                handled_vars.add(ds_info["name"])
                yaml_info = ds_info
            if ds_info["name"] == "BT":
                yield from self._available_btemp_datasets(yaml_info)
            yield True, ds_info
        yield from self._available_new_datasets(handled_vars)

    def _count_channel_repeat_number(self):
        """Count channel/polarization pair repetition."""
        freq = self.nc.coords.get("Freq", self.nc.get("Freq"))
        polo = self.nc["Polo"]

        chn_total = Counter()
        normals = []
        for idx, (f, p) in enumerate(zip(freq, polo)):
            normal_f = str(int(f))
            normal_p = "v" if p == POLO_V else "h"
            chn_total[normal_f + normal_p] += 1
            normals.append((idx, f, p, normal_f, normal_p))

        return chn_total, normals

    def _available_btemp_datasets(self, yaml_info):
        """Create metadata for channel BTs."""
        chn_total, normals = self._count_channel_repeat_number()
        # keep track of current channel count for string description
        chn_cnt = Counter()
        for idx, _f, _p, normal_f, normal_p in normals:
            chn_cnt[normal_f + normal_p] += 1
            p_count = str(chn_cnt[normal_f + normal_p]
                          if chn_total[normal_f + normal_p] > 1 else "")

            new_name = "btemp_{}{}{}".format(normal_f, normal_p, p_count)

            desc_bt = "Channel {} Brightness Temperature at {}GHz {}{}"
            desc_bt = desc_bt.format(idx, normal_f, normal_p, p_count)
            ds_info = yaml_info.copy()
            ds_info.update({
                "file_type": self.filetype_info["file_type"],
                "name": new_name,
                "description": desc_bt,
                "channel_index": idx,
                "frequency": "{}GHz".format(normal_f),
                "polarization": normal_p,
                "dependencies": ("BT", "Sfc_type"),
                "coordinates": ["longitude", "latitude"]
            })
            yield True, ds_info

    def _get_ds_info_for_data_arr(self, var_name):
        ds_info = {
            "file_type": self.filetype_info["file_type"],
            "name": var_name,
            "coordinates": ["longitude", "latitude"]
        }
        return ds_info

    def _is_2d_yx_data_array(self, data_arr):
        has_y_dim = data_arr.dims[0] == "y"
        has_x_dim = data_arr.dims[1] == "x"
        return has_y_dim and has_x_dim

    def _available_new_datasets(self, handled_vars):
        """Metadata for available variables other than BT."""
        possible_vars = list(self.nc.items()) + list(self.nc.coords.items())
        for var_name, data_arr in possible_vars:
            if var_name in handled_vars:
                continue
            if data_arr.ndim != 2:
                # we don't currently handle non-2D variables
                continue
            if not self._is_2d_yx_data_array(data_arr):
                # we need 'traditional' y/x dimensions currently
                continue

            ds_info = self._get_ds_info_for_data_arr(var_name)
            yield True, ds_info

    def __getitem__(self, item):
        """Wrap around `self.nc[item]`."""
        data = self.nc[item]

        # 'Freq' dimension causes issues in other processing
        if "Freq" in data.coords:
            data = data.drop_vars("Freq")

        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for MODIS L1b HDF reader."""

from __future__ import annotations

import dask
import numpy as np
import pytest
from pytest_lazy_fixtures import lf as lazy_fixture

from satpy import Scene, available_readers
from satpy.tests.utils import CustomScheduler, make_dataid

from ._modis_fixtures import (
    AVAILABLE_1KM_PRODUCT_NAMES,
    AVAILABLE_HKM_PRODUCT_NAMES,
    AVAILABLE_QKM_PRODUCT_NAMES,
    _shape_for_resolution,
)

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - modis_l1b_imapp_1000m_file
# - modis_l1b_imapp_geo_file,
# - modis_l1b_nasa_1km_mod03_files,
# - modis_l1b_nasa_mod02hkm_file,
# - modis_l1b_nasa_mod02qkm_file,
# - modis_l1b_nasa_mod03_file,
# - modis_l1b_nasa_mod021km_file


def _check_shared_metadata(data_arr):
    assert data_arr.attrs["sensor"] == "modis"
    assert data_arr.attrs["platform_name"] == "Terra"
    assert "rows_per_scan" in data_arr.attrs
    assert isinstance(data_arr.attrs["rows_per_scan"], int)
    assert data_arr.attrs["reader"] == "modis_l1b"
    assert "resolution" in data_arr.attrs
    res = data_arr.attrs["resolution"]
    if res == 5000:
        assert data_arr.chunks == ((2, 2, 2), (data_arr.shape[1],))
    elif res == 1000:
        assert data_arr.chunks == ((10, 10, 10), (data_arr.shape[1],))
    elif res == 500:
        assert data_arr.chunks == ((20, 20, 20), (data_arr.shape[1],))
    elif res == 250:
        assert data_arr.chunks == ((40, 40, 40), (data_arr.shape[1],))
    else:
        raise ValueError(f"Unexpected resolution: {res}")


def _load_and_check_geolocation(scene, resolution, exp_res, exp_shape, has_res,
                                check_callback=_check_shared_metadata):
    scene.load(["longitude", "latitude"], resolution=resolution)
    lon_id = make_dataid(name="longitude", resolution=exp_res)
    lat_id = make_dataid(name="latitude", resolution=exp_res)
    if has_res:
        lon_arr = scene[lon_id]
        lat_arr = scene[lat_id]
        assert lon_arr.shape == exp_shape
        assert lat_arr.shape == exp_shape
        # compute lon/lat at the same time to avoid wasted computation
        lon_vals, lat_vals = dask.compute(lon_arr, lat_arr)
        assert lon_arr.dtype == lat_arr.dtype
        assert lon_arr.dtype == np.float32
        assert lon_vals.dtype == lon_arr.dtype
        assert lat_vals.dtype == lat_arr.dtype
        np.testing.assert_array_less(lon_vals, 0)
        np.testing.assert_array_less(0, lat_vals)
        check_callback(lon_arr)
        check_callback(lat_arr)
    else:
        pytest.raises(KeyError, scene.__getitem__, lon_id)
        pytest.raises(KeyError, scene.__getitem__, lat_id)


class TestModisL1b:
    """Test MODIS L1b reader."""

    def test_available_reader(self):
        """Test that MODIS L1b reader is available."""
        assert "modis_l1b" in available_readers()

    @pytest.mark.parametrize(
        ("input_files", "expected_names", "expected_data_res", "expected_geo_res"),
        [
            (lazy_fixture("modis_l1b_nasa_mod021km_file"),
             AVAILABLE_1KM_PRODUCT_NAMES + AVAILABLE_HKM_PRODUCT_NAMES + AVAILABLE_QKM_PRODUCT_NAMES,
             [1000], [5000, 1000]),
            (lazy_fixture("modis_l1b_imapp_1000m_file"),
             AVAILABLE_1KM_PRODUCT_NAMES + AVAILABLE_HKM_PRODUCT_NAMES + AVAILABLE_QKM_PRODUCT_NAMES,
             [1000], [5000, 1000]),
            (lazy_fixture("modis_l1b_nasa_mod02hkm_file"),
             AVAILABLE_HKM_PRODUCT_NAMES + AVAILABLE_QKM_PRODUCT_NAMES, [500], [1000, 500, 250]),
        (lazy_fixture("modis_l1b_nasa_mod02qkm_file"),
             AVAILABLE_QKM_PRODUCT_NAMES, [250], [1000, 500, 250]),
        ]
    )
    def test_scene_available_datasets(self, input_files, expected_names, expected_data_res, expected_geo_res):
        """Test that datasets are available."""
        scene = Scene(reader="modis_l1b", filenames=input_files)
        available_datasets = scene.available_dataset_names()
        assert len(available_datasets) > 0
        assert "longitude" in available_datasets
        assert "latitude" in available_datasets
        for chan_name in expected_names:
            assert chan_name in available_datasets

        available_data_ids = scene.available_dataset_ids()
        available_datas = {x: [] for x in expected_data_res}
        available_geos = {x: [] for x in expected_geo_res}
        # Make sure that every resolution from the reader is what we expect
        for data_id in available_data_ids:
            res = data_id["resolution"]
            if data_id["name"] in ["longitude", "latitude"]:
                assert res in expected_geo_res
                available_geos[res].append(data_id)
            else:
                assert res in expected_data_res
                available_datas[res].append(data_id)

        # Make sure that every resolution we expect has at least one dataset
        for exp_res, avail_id in available_datas.items():
            assert avail_id, f"Missing datasets for data resolution {exp_res}"
        for exp_res, avail_id in available_geos.items():
            assert avail_id, f"Missing geo datasets for geo resolution {exp_res}"

    @pytest.mark.parametrize(
        ("input_files", "has_5km", "has_500", "has_250", "default_res"),
        [
            (lazy_fixture("modis_l1b_nasa_mod021km_file"),
             True, False, False, 1000),
            (lazy_fixture("modis_l1b_imapp_1000m_file"),
             True, False, False, 1000),
            (lazy_fixture("modis_l1b_nasa_mod02hkm_file"),
             False, True, True, 250),
            (lazy_fixture("modis_l1b_nasa_mod02qkm_file"),
             False, True, True, 250),
            (lazy_fixture("modis_l1b_nasa_1km_mod03_files"),
             True, True, True, 250),
        ]
    )
    def test_load_longitude_latitude(self, input_files, has_5km, has_500, has_250, default_res):
        """Test that longitude and latitude datasets are loaded correctly."""
        scene = Scene(reader="modis_l1b", filenames=input_files)
        shape_5km = _shape_for_resolution(5000)
        shape_500m = _shape_for_resolution(500)
        shape_250m = _shape_for_resolution(250)
        default_shape = _shape_for_resolution(default_res)
        scheduler = CustomScheduler(max_computes=1 + has_5km + has_500 + has_250)
        with dask.config.set({"scheduler": scheduler, "array.chunk-size": "1 MiB"}):
            _load_and_check_geolocation(scene, "*", default_res, default_shape, True)
            _load_and_check_geolocation(scene, 5000, 5000, shape_5km, has_5km)
            _load_and_check_geolocation(scene, 500, 500, shape_500m, has_500)
            _load_and_check_geolocation(scene, 250, 250, shape_250m, has_250)

    def test_load_sat_zenith_angle(self, modis_l1b_nasa_mod021km_file):
        """Test loading satellite zenith angle band."""
        scene = Scene(reader="modis_l1b", filenames=modis_l1b_nasa_mod021km_file)
        dataset_name = "satellite_zenith_angle"
        with dask.config.set({"array.chunk-size": "1 MiB"}):
            scene.load([dataset_name])
        dataset = scene[dataset_name]
        assert dataset.shape == _shape_for_resolution(1000)
        assert dataset.attrs["resolution"] == 1000
        _check_shared_metadata(dataset)

    def test_load_vis(self, modis_l1b_nasa_mod021km_file):
        """Test loading visible band."""
        scene = Scene(reader="modis_l1b", filenames=modis_l1b_nasa_mod021km_file)
        dataset_name = "1"
        with dask.config.set({"array.chunk-size": "1 MiB"}):
            scene.load([dataset_name])
        dataset = scene[dataset_name]
        assert dataset[0, 0] == 300.0
        assert dataset.shape == _shape_for_resolution(1000)
        assert dataset.attrs["resolution"] == 1000
        _check_shared_metadata(dataset)

    @pytest.mark.parametrize("mask_saturated", [False, True])
    def test_load_vis_saturation(self, mask_saturated, modis_l1b_nasa_mod021km_file):
        """Test loading visible band."""
        scene = Scene(reader="modis_l1b", filenames=modis_l1b_nasa_mod021km_file,
                      reader_kwargs={"mask_saturated": mask_saturated})
        dataset_name = "2"
        with dask.config.set({"array.chunk-size": "1 MiB"}):
            scene.load([dataset_name])
        dataset = scene[dataset_name]
        assert dataset.shape == _shape_for_resolution(1000)
        assert dataset.attrs["resolution"] == 1000
        _check_shared_metadata(dataset)

        # check saturation fill values
        data = dataset.values
        assert dataset[0, 0] == 300.0
        assert np.isnan(data[-1, -1])  # normal fill value
        if mask_saturated:
            assert np.isnan(data[-1, -2])  # saturation
            assert np.isnan(data[-1, -3])  # can't aggregate
        else:
            # test data factor/offset are 1/0
            # albedos are converted to %
            assert data[-1, -2] >= 32767 * 100.0  # saturation
            assert data[-1, -3] >= 32767 * 100.0  # can't aggregate
